/*
Copyright  1993 Bengt Mrtensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

// $Id: CIntegrator.cc,v 1.7 1994/08/24 09:58:06 bengt Exp $

#if __GNUC__
#pragma implementation
#endif

#include "externals.h"
#include "CIntegrator.h"

std::string ContinuousIntegrator::xerrwv_themessage =  "";
bool ContinuousIntegrator::xerrwv_initialized = true;

// These functions are (somewhat) commented in CIntegrator.h

void				// Override me in riccati and stoch and ...
ContinuousIntegrator::initialize_integration()
{
  junk_vector.resize(current_system.number_of_states);
  (*current_system.f_ptr)(systemtime, current_system.state_vector,
			  junk_vector);
  (*current_system.f_store_ptr)();
  (*current_system.f_accepted_ptr)();
}

void
ContinuousIntegrator::hairy_f(const int */*n*/, double *t, double *x, double *dx)
{
  (*current_system.f_ptr)(t, x, dx);
}

bool
ContinuousIntegrator::finished() const
{
  const double very_small = 1e-8;
  return
    (signed_step == 0
     ? (stoptime - systemtime)/(stoptime - starttime) < very_small
     : (starttime < stoptime
	? systemtime > stoptime - signed_step
	: systemtime < stoptime - signed_step))
    || ! running_integration;
}

bool
ContinuousIntegrator::more_to_do_zero_step() const
{
  return (starttime < stoptime ? systemtime < stoptime : systemtime > stoptime)
	  && running_integration && ! has_been_alarmed;
}

bool
ContinuousIntegrator::continuous_time_barf()
{
  if (current_system.system_type == discrete
      || current_system.system_type == riccati)
    // Programming error
    gnans_error(impossible, "Ooops, continuous algorithm selected.");
  return false;
}

bool
ContinuousIntegrator::barf_if_step_zero()
{
  if (signed_step == 0) {
    gnans_error(non_fatal,
                "The selected algorithm does not allow zero stepsize");
    return true;
  }
  return false;
}

bool
ContinuousIntegrator::barf_on_wieners()
{
  if (current_system.number_of_wieners && !expert_mode) {
    gnans_error(non_fatal,
                "The selected algorith does not support Wiener processes");
    return true;
  }
  return false;
}

bool
ContinuousIntegrator::overshot() const
{
  return starttime < stoptime ? systemtime > stoptime : systemtime < stoptime;
}

void
ContinuousIntegrator::integration_error(double t1, double t2,
					signed int errorcode,
					const integration_error_message
					   *errors,
					const char *msg /*= NULL*/) const
{
  const char *message = "*** No error description found ***";
  for (unsigned int i = 0; errors[i].get_errorno(); i++)
    if (errors[i].get_errorno() == errorcode) {
      message = errors[i].get_error_message();
      break;
    }
  if (msg)
    gnans_error(warning,
                "Integration error #%i between time %g and time %g\n"
                "Error description: %s\n"
                "Messages from integration algorithm:\n"
                "================================\n%s",
                errorcode, t1, t2, message, msg);
  else
    gnans_error(warning,
                "Integration error #%i between time %g and time %g\n"
                "Error description: %s",
                errorcode, t1, t2, message);
}

// This is a generic driver routine for a dumb, fixed stepsize algorithm
// without  error control.
bool
ContinuousIntegrator::loop(bool initialize)
{
  // 1. Initialization
  if (initialize) {
    bool success = initialize_integrator();
    if (! success)
      return true;
  }

  // 2. The inner loop
  do {
    // 2.1 Conditionally store
    store_if_necessary();
    
    // 2.2 Call the integrator
    call_integrator();
    (*current_system.f_accepted_ptr)();
  } while (more_to_do_nonzero_step());
  
  // Got here either if finished, the integration was stopped, or alarm
  
  // 3. Unconditionally store if finished
  if (finished())
    store_stuff();

  // 4. If I got here due to the alarm, return false. Else return true.
  return finished();
}

// Returns what has previously been "written" 
std::string
ContinuousIntegrator::xerrwv_flush()
{
  if (!ContinuousIntegrator::xerrwv_initialized) {
    ContinuousIntegrator::xerrwv_themessage =  "";
    ContinuousIntegrator::xerrwv_initialized = true;
  }
  std::string s(ContinuousIntegrator::xerrwv_themessage);
  ContinuousIntegrator::xerrwv_themessage = "";
#if 0
  if (s.find("\n"))
    s.at("\n", -1) = "";
#endif
  return s;
}
