/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

// $Id: $

#ifdef __GNUG__
#pragma implementation
#endif

#include "externals.h"
#include "utillib.h"
#include "HIntegrator.h"

static integration_error_message Hindmarsh_errors[] = {
  { -1, "Excess work done on this call"},
  { -2, "Excess accuracy requested"},
  { -3, "Illegal input detected"},
  { -4, "Repeated error test failures"},
  { -5, "Repeated convergence failures"},
  { -6, "Error weight became zero during problem"},
  { 0, ""}
};

HindmarshIntegrator::HindmarshIntegrator()
{
  neqn = -1;			// Signals impossible condition, to restart
  itol = 1;
  iopt = 1;
  liw =  0;
  rwork = NULL;
  iwork = NULL;
}


// The driver for Hindmarsh. The structure is similar to that of the Euler
// driver. Check the comments in the Euler driver.
bool
HindmarshIntegrator::loop(bool initialize)
{
  if (initialize) {
    bool success = initialize_integrator();
    if (! success)
      return true;
  }

  // If something went wrong previously, tell the algo to re-initialize
  if (istate < 0)
    istate = 1;
  
  // Double the code for storeincrement > 0
  // If timestep = 0 we let the algorithm chose its own steps.
  // For efficiency, we double the code for that case too.
  if (storeincrement == 0.0) {
    /***************** Store everything *****************/
    if (signed_step != 0) {
      /***************** Prescribed stepsize ****************/
      itask = 1;
      do {
	store_stuff();
	tout = systemtime + signed_step;
	call_integrator();
	(*current_system.f_accepted_ptr)();
      } while (more_to_do_nonzero_step() && istate == 2);
    } else {
      /***************** Free stepsize ****************/
      tout = stoptime;
      itask = 2;
      do {
	store_stuff();
	call_integrator();
	(*current_system.f_accepted_ptr)();
      } while (more_to_do_zero_step() && istate == 2);
    }
  } else {
    /***************** Store selectively *****************/
    if (signed_step != 0) {
      /***************** Prescribed stepsize ****************/
      itask = 1;
      do {
	store_if_necessary();
	tout = systemtime + signed_step;
	call_integrator();
	(*current_system.f_accepted_ptr)();
      } while (more_to_do_nonzero_step() && istate == 2);
    } else {
      /***************** Free stepsize ****************/
      tout = stoptime;
      itask = 2;
      do {
	store_if_necessary();
	call_integrator();
	(*current_system.f_accepted_ptr)();
      } while (more_to_do_zero_step() && istate == 2);
    }
  }
  
  if (finished()) {
    if (overshot()) {
      // Overshot, don't store but go back to stoptime
      itask = 1;
      tout = stoptime;
      call_integrator();
    }
    store_stuff();
  }
  if (istate != 2) {
    integration_error(systemtime, tout,  istate, Hindmarsh_errors, 
                      xerrwv_flush().c_str());
    return true;
  }
  return finished();
}

bool
HindmarshIntegrator::initialize_integrator()
{
  setup_times();
  if (barf_on_wieners() || continuous_time_barf())
    return false;
    
  if (int(method_flag) != 10 && int(method_flag) != 22) {
    gnans_error(non_fatal, "Unsupported method_flag (= %d).",
                int(method_flag));
    return false;
  }

  int oldmf = mf;
  mf = int(method_flag);
  istate = 1;

  make_setup(oldmf);
                        
  // Set up some smarter defaults for the integrator.
  // We write iwork[k-1] to never forget that Fortran's vectors start
  // with index 1.
  for (unsigned int i = 5; i <= 10; i++) {
    iwork[i-1] = 0;
    rwork[i-1] = 0;
  }
    
  // Don't ever take larger step than whole interval/10
  rwork[6-1] = abs(stoptime - starttime)/10;

  // Min stepsize
  // rwork[7-1] = abs(stoptime - starttime)/1e12;

  // Issue warning that t + h = t on a step (h = step size) only once.
  iwork[7-1] = 1;

  initialize_integration();
  return true;
}

