/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

// $Id: Newton.cc,v 1.3 1994/08/19 18:21:46 bengt Exp $

#ifdef __GNUG__
#pragma implementation
#endif

#include "Newton.h"

const char *
NewtonIntegrator::name() const
{
  return "Newton";
}

const char *
NewtonIntegrator::documentation() const
{
  return
    "Explicit, asymptotically efficient, stochastic Runge-Kutta type method\n"
    "for SDE's with only one noise source.\n"
    "Converges with strong order 1.0 to the It solution.\n"
    "Fixed step-size, no error control.\n\n"
    "Reference:\n"
    "   Nigel J. Newton: \"Asymptotically Efficient Runge-Kutta Methods For\n"
    "   A Class of It and Stratonovich Equations\", SIAM J. Appl. Math.\n"
    "   Vol 51, No 2, pp. 542-567, April 1991";
}

// This is the integator.
// Note: The algorithm is derived for time invariant vector fields.
// If the vector fields are time varying, then...
inline void
NewtonIntegrator::newton(fp f, unsigned int /*number_of_states*/,
			 double *tp, double h, Vect &y)
{
  f(tp, y, a0);
  b0 = current_system.wiener_coeffs[0];
  
  build_wiener_table(wiener_increments);
  double delta_W = sqrt_step()*((double*)(wiener_increments))[0];

  // y1 = y + a0*h + b0*delta_W;
  y1 = y;
  y2 = a0;
  y2 *= h;
  y1 += y2;
  y2 = b0;
  y2 *= delta_W;
  y1 += y2;

  f(tp, y1, a1);

  // y1 = y - q_2_3*b0*(delta_W + sqrt_3_h)
  y1 = y;
  y2 = b0;
  y2 *= q_2_3*(delta_W + sqrt_3_h);
  y1 -= y2;
  
  f(tp, y1, scratch);
  b1 = current_system.wiener_coeffs[0];

  // y1 = y + q_2_9*b0*(3*delta_W + sqrt_3_h);
  y1 = y;
  y2 = b0;
  y2 *= q_2_9*(3*delta_W + sqrt_3_h);
  y1 += y2;
  f(tp, y1, scratch);
  b2 = current_system.wiener_coeffs[0];

  // y1 = y - q_20_27*a0*h + q_10_27*(b1 - b0)*delta_W - q_10_27*b1*sqrt_3_h;
  y1 = y;
  y2 = a0;
  y2 *= q_20_27*h;
  y1 += y2;
  y2 = b1;
  y2 -= b0;
  y2 *= q_10_27*delta_W;
  y1 += y2;
  y2 = b1;
  y2 *= q_10_27*sqrt_3_h;
  y1 -= y2;

  f(tp, y1, scratch);
  b3 = current_system.wiener_coeffs[0];

  //  y += (q_h_2*(a0 + a1)
  //	+ q_1_40*(37*b0 + 30*b2 - 27*b3)*delta_W
  //	+ q_1_16*sqrt_3_h*(8*b0 + b1 - 9*b2));
  y1 = a0;
  y1 += a1;
  y1 *= q_h_2;
  y += y1;

  y2 = b0;
  y2 *= 37;
  y1 = y2;
  y2 = b2;
  y2 *=30;
  y1 += y2;
  y2 = b3;
  y2 *= 27;
  y1 -= y2;
  y1 *= q_1_40*delta_W;
  y += y1;

  y2 = b0;
  y2 *= 8;
  y1 = y2;
  y1 += b1;
  y2 = b2;
  y2 *= 9;
  y1 -= y2;
  y1 *= q_1_16*sqrt_3_h;
  y += y1;    

  *tp += h;
}

inline bool
NewtonIntegrator::call_integrator()
{
  newton(current_system.f_ptr, current_system.number_of_states,
	 systemtime, timestep, current_system.state_vector);
  return true;
}

bool
NewtonIntegrator::initialize_integrator()
{
  q_2_3 = 2.0/3.0;
  q_2_9 = 2.0/9.0;
  q_20_27 = 20.0/27.0;
  q_10_27 = 10.0/27.0;
  q_1_40 = 1.0/40.0;
  q_1_16 = 1.0/16.0;
   
  // Assigning initial values to some internal variables
  setup_times();
    
  // Algorithm specific checks
  if (continuous_time_barf() || barf_if_step_zero() || barf_if_backwards()
      || scalar_noise_barf())
    return false;

  // Algorithm independent initializations
  initialize_integration();

  // Reserve space for arrays etc.
  scratch.resize(current_system.number_of_states);
  y1.resize(current_system.number_of_states);
  y2.resize(current_system.number_of_states);

  a0.resize(current_system.number_of_states);
  a1.resize(current_system.number_of_states);

  b0.resize(current_system.number_of_states);
  b1.resize(current_system.number_of_states);
  b2.resize(current_system.number_of_states);
  b3.resize(current_system.number_of_states);

  sqrt_3_h = sqrt(3*double(timestep));
  q_h_2 = double(timestep)/2;
  
  return true;
}
