/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

// $Id: Ode.cc,v 1.5 1994/08/19 18:21:46 bengt Exp $

#ifdef __GNUG__
#pragma implementation
#endif

#include "externals.h"
#include "Ode.h"

#ifdef FORTRAN_UNDERSCORES
#define ode ode_
#endif

#ifdef FORTRAN_UPPERCASE
#define ode ODE
#endif

extern "C" int 
ode(fp f, int *neqn, double *y, double *t,
    double *tout, double *relerr, double *abserr,
    int *iflag, double *work, int *iwork);

const char *
OdeIntegrator::name() const
{
  return "Ode";
}

const char *
OdeIntegrator::documentation() const
{
  return
    "Adams method IVP from Shampine and Gordon book.\n"
    "\n"
    "Uses a modified divided difference form of the Adams Pece\n"
    "formulas and local extrapolation.  It adjusts the order and step\n"
    "size to control the local error per unit step in a generalized\n"
    "sense.  Normally each call to  step  advances the solution one step\n"
    "in the direction of  tout.  For reasons of efficiency  de\n"
    "integrates beyond  tout  internally, though never beyond\n"
    "t+10*(tout-t), and calls intrp to interpolate the solution at\n"
    "tout.\n"
    "\n"
    "This code is completely explained and documented in the text:\n"
    "\n"
    "Computer Solution Of Ordinary Differential Equations:  The Initial\n"
    "Value Problem  By L. F. Shampine and M. K. Gordon.";
}

static integration_error_message ode_errors[] = {
  { 3, "Error tolerances too small. Relerr, Abserr increased\n"
      "appropriately for continuing"},
  { 4, "More than 500 steps needed"},
  { 5, "Equations appear to be stiff"},
  { 6, "Invalid input parameters (fatal error)"},
  { 0, ""}
};

inline bool
OdeIntegrator::call_integrator()
{
  ode(current_system.f_ptr, &neqn, current_system.state_vector,
      systemtime, &tout, relerror, abserror, &iflag, work, iwork);
  return true;
}

OdeIntegrator::OdeIntegrator()
{
  neqn = -1;
  work = NULL;
}

bool
OdeIntegrator::initialize_integrator()
{
  setup_times();
  if (continuous_time_barf() || barf_if_step_zero() || barf_on_wieners())
    return false;

  if (unsigned(neqn) != current_system.number_of_states) {
    neqn =  current_system.number_of_states;
    delete [] work;
    work = new double [100 + 21*current_system.number_of_states];
  }

  iflag = 1;
  initialize_integration();
  return true;
}

bool
OdeIntegrator::loop(bool initialize)
{
  if (initialize) {
    bool success = initialize_integrator();
    if (! success)
      return true;
  }
  
  if (storeincrement == 0.0) {
    // Store everything
    do {
      store_stuff();
      tout = systemtime + signed_step;
      call_integrator();
      (*current_system.f_accepted_ptr)();
    } while (more_to_do_nonzero_step() && iflag == 2);
  } else {
    // Store selectively
    do {
      store_if_necessary();
      tout = systemtime + signed_step;
      call_integrator();
      (*current_system.f_accepted_ptr)();
    } while (more_to_do_nonzero_step() && iflag == 2);
  }
  
  if (finished()) {
    if (overshot()) {
      // Overshot, don't store but go back to stoptime
      tout = stoptime;
      call_integrator();
    }
    store_stuff();
  }
  
  if (iflag != 2) {
    integration_error(systemtime, tout, iflag, ode_errors);
    return true;
  }
  return finished();
}
