/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

// $Id: Rkf45.cc,v 1.5 1994/08/19 18:21:46 bengt Exp $

#ifdef __GNUG__
#pragma implementation
#endif

#include "externals.h"
#include "utillib.h"
#include "Rkf45.h"

#ifdef FORTRAN_UNDERSCORES
#define rkf45 rkf45_
#endif

#ifdef FORTRAN_UPPERCASE
#define rkf45 RKF45
#endif

// External declaration for the integrator
extern "C" void 
rkf45(fp f, int *neqn, double *y, double *t,
      double *tout, double *relerr, double *abserr,
      int *iflag, double *work, int *iwork);

const char *
Rkf45Integrator::name() const
{
  return "RKF45";
}

const char *
Rkf45Integrator::documentation() const
{
  return
    "Watts and Shampine Runge-Kutta Fehlberg 45 IVP solver\n"
    "\n"
    "Fehlberg fourth-fifth order Runge-Kutta method\n"
    "\n"
    "    Written by H. A. Watts and L. F. Shampine\n"
    "                  Sandia Laboratories\n"
    "                  Albuquerque, New Mexico\n"
    "\n"
    "RKF45 is primarily designed to solve non-stiff and mildly stiff\n"
    "differential equations when derivative evaluations are inexpensive.\n"
    "RKF45 should generally not be used when the user is demanding\n"
    "high accuracy.\n"
    "\n"
    "RKF45  uses the Runge-Kutta-Fehlberg (4,5)  method described\n"
    "in the reference\n"
    "   E. Fehlberg, Low-Order Classical Runge-Kutta Formulas With Stepsize\n"
    "                Control, NASA TR R-315\n"
    "\n"
    "The performance of RKF45 is illustrated in the reference\n"
    "\n"
    "	L. F. Shampine, H. A. Watts, S. Davenport, Solving Non-Stiff Ordinary\n"
    "        Differential Equations--The State Of The Art,\n"
    "        Sandia Laboratories Report SAND75-0182,\n"
    "        (To appear in SIAM Review.)";
}

static integration_error_message rkf45_errors[] = {
  { 3, "Integration was not completed because relative error\n"
      "tolerance was too small. Relerr has been increased\n"
      "appropriately for continuing"},
  { 4, "Integration was not completed because more than\n"
      "3000 derivative evaluations were needed. This\n"
      "is approximately 500 steps"},
  { 5, "Integration was not completed because solution\n"
      "vanished making a pure relative error test\n"
      "impossible. Must use non-zero abserr to continue.\n"
      "Using the one-step integration mode for one step\n"
      "is a good way to proceed."},
  { 6, "Integration was not completed because requested\n"
      "accuracy could not be achieved using smallest\n"
      "allowable stepsize. User must increase the error\n"
      "tolerance before continued integration can be\n"
      "attempted."},
  { 7, "It is likely that RKF45 is inefficient for solving\n"
      "this problem. Too much output is restricting the\n"
      "natural stepsize choice. Use the one-step integrator\n"
      "mode."},
  { 8, "Invalid input parameters (Probably internal error)"},
  {0, ""}
};

inline bool
Rkf45Integrator::call_integrator()
{
  rkf45(current_system.f_ptr,
	&neqn,
	current_system.state_vector,
	systemtime,
	&tout,
	relerror,
	abserror,
	&iflag,
	work,
	iwork);
  return true;
}

Rkf45Integrator::Rkf45Integrator()
{
  neqn = -1;
  work = NULL; 
}

bool
Rkf45Integrator::initialize_integrator()
{
  setup_times();
  if (continuous_time_barf() || barf_on_wieners())
    return false;

  if (unsigned(neqn) != current_system.number_of_states) {
    neqn = current_system.number_of_states;
    delete [] work;
    work = new double[3 + 6*current_system.number_of_states];
  }
  iflag = 1;
    
  initialize_integration();
  return true;
}

bool
Rkf45Integrator::loop(bool initialize)
{
  if (initialize) {
    bool success = initialize_integrator();
    if (! success)
      return true;
  }
   
  // Double the code for storeincrement > 0
  // If timestep = 0 we let the algorithm chose its own steps.
  // For efficiency, we double the code for that case too.
  if (storeincrement == 0.0) {
    // Store everything!
    if (signed_step != 0)
      // Prescribed stepsize
      do {
        store_stuff();
        tout = systemtime + signed_step;
        call_integrator();
	(*current_system.f_accepted_ptr)();
      } while (more_to_do_nonzero_step() && iflag == 2);
    else {
      // Free stepsize
      tout = stoptime;
      do {
        store_stuff();
        iflag = -1;
        call_integrator();
	(*current_system.f_accepted_ptr)();
      } while (more_to_do_zero_step() && iflag == -2);
    }
  } else {
    // Store selectively
    if (signed_step != 0)
      // Prescribed stepsize
      do {
        store_if_necessary();
        tout = systemtime + signed_step;
        call_integrator();
	(*current_system.f_accepted_ptr)();
      } while (more_to_do_nonzero_step() && iflag == 2);
    else {
      // Free stepsize
      tout = stoptime;
      do {
        store_if_necessary();
        iflag = -1;
        call_integrator();
	(*current_system.f_accepted_ptr)();
      } while (more_to_do_zero_step() && iflag == -2);
    }
  }
  
  if (finished())
    store_stuff();
  if (abs(iflag) != 2) {
    integration_error(systemtime, tout,  iflag, rkf45_errors);
    return true;
  }
  return finished();
}
