/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

// $Id: $

#ifdef __GNUG__
#pragma implementation
#endif

#include "externals.h"
#include "utillib.h"
#include "Vode.h"

#ifdef FORTRAN_UNDERSCORES
#define dvode dvode_
#endif

#ifdef FORTRAN_UPPERCASE
#define dvode DVODE
#endif

// External declaration for the integrator
extern "C" void 
#ifdef __GNUG__
dvode(hairy_fp f, int *neq, double *y, double *t,
      double *tout, int *itol, double *rtol,
      double *atol, int *itask, int *istate,
      int *iopt, double *rwork, int *lrw,
      int *iwork, int *liw, void *jac(), int *mf);
#else
dvode(void (*)(const int *, double *, double *, double *), int *, double *, double *,
      double *, int *, double *,
      double *, int *, int *,
      int *, double *, int *,
      int *, int *, void (*)(), int *);
#endif

const char *
VodeIntegrator::documentation() const
{
  return
    " VODE: Variable-coefficient Ordinary Differential Equation solver,\n"
    " with fixed-leading coefficient implementation.\n"
    "\n"
    " VODE solves the initial value problem for stiff or nonstiff\n"
    " systems of first order ODEs,\n"
    "     dy/dt = f(t,y) ,  or, in component form,\n"
    "     dy(i)/dt = f(i) = f(i,t,y(1),y(2),...,y(NEQ)) (i = 1,...,NEQ).\n"
    " VODE is a package based on the EPISODE and EPISODEB packages, and\n"
    " on the ODEPACK user interface standard, with minor modifications.\n"
    "--------------------------------------------------\n"
    " References:\n"
    "\n"
    " 1. P. N. Brown, G. D. Byrne, and A. C. Hindmarsh, VODE: A Variable\n"
    "    Coefficient ODE Solver, SIAM J. Sci. Stat. Comput., 10 (1989),\n"
    "    pp. 1038-1051.  Also, LLNL Report UCRL-98412, June 1988.\n"
    " 2. G. D. Byrne and A. C. Hindmarsh, A Polyalgorithm for the\n"
    "    Numerical Solution of Ordinary Differential Equations,\n"
    "    ACM Trans. Math. Software, 1 (1975), pp. 71-96.\n"
    " 3. A. C. Hindmarsh and G. D. Byrne, EPISODE: An Effective Package\n"
    "    for the Integration of Systems of Ordinary Differential\n"
    "    Equations, LLNL Report UCID-30112, Rev. 1, April 1977.\n"
    " 4. G. D. Byrne and A. C. Hindmarsh, EPISODEB: An Experimental\n"
    "    Package for the Integration of Systems of Ordinary Differential\n"
    "    Equations with Banded Jacobians, LLNL Report UCID-30132, April\n"
    "    1976.\n"
    " 5. A. C. Hindmarsh, ODEPACK, a Systematized Collection of ODE\n"
    "    Solvers, in Scientific Computing, R. S. Stepleman et al., eds.,\n"
    "    North-Holland, Amsterdam, 1983, pp. 55-64.\n"
    " 6. K. R. Jackson and R. Sacks-Davis, An Alternative Implementation\n"
    "    of Variable Step-Size Multistep Formulas for Stiff ODEs, ACM\n"
    "    Trans. Math. Software, 6 (1980), pp. 295-318.\n";
}

// Define inline function to guarantee consistency
//inline 
bool
VodeIntegrator::call_integrator()
{
  dvode(&hairy_f, &neqn, current_system.state_vector,
	systemtime, &tout, &itol, relerror, abserror,
	&itask, &istate, &iopt, rwork, &lrw, iwork, &liw,
	NULL, &mf);
  return true;
}

bool
VodeIntegrator::make_setup(int oldmf)
{
  if (unsigned(neqn) != current_system.number_of_states || oldmf != mf) {
    neqn = current_system.number_of_states;
    lrw = mf == 10 ? 20 + 16*neqn :
          mf == 22 ? 22 +  9*neqn + 2*neqn*neqn : 0;
    delete [] rwork;
    rwork = new double[(unsigned) lrw];
    delete [] iwork;
    liw = 30 + (mf != 10)*neqn;
    iwork = new int[liw];
  }
  return rwork && iwork;
}
