/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

// $Id: Plotter.cc,v 1.8 1994/11/01 12:52:38 bengt Exp bengt $

// This file is based on work by Patrick Rosenberg 

#ifdef __GNUG__
#pragma implementation
#endif

#include "externals.h"
#include <sys/types.h>
#include <unistd.h>

#include <X11/Intrinsic.h>
#include <X11/bitmaps/xlogo16>

typedef char *String;		// FIXME
extern "C" {
#include <X11/At/Plotter.h>	// Athena Plotter Widget Set V6.0
}

static void PlotType(Widget, XtPointer, XtPointer);
static void Zoom(Widget, XtPointer, XtPointer);
static void Zooming(Widget w, XtPointer client_data,
		    AtRectangleCallbackData *cad);
static void Grid(Widget, XtPointer, XtPointer);
static void Subgrid(Widget, XtPointer, XtPointer);
static void PrintPostscript(Widget, XtPointer, XtPointer);
static void Linlog_select(Widget, XtPointer, XtPointer);
static void DeletePlotter(Widget, XtPointer, XtPointer);

#include "Plotter.h"
#include "PlotterControl.h"

#ifdef USEXAW3DINCLUDES

//#include <X11/Xaw3d/Box.h>
#include <X11/Xaw3d/Command.h>
#include <X11/Xaw3d/Form.h>
#include <X11/Xaw3d/MenuButton.h>
#include <X11/Xaw3d/SimpleMenu.h>
#include <X11/Xaw3d/SmeBSB.h>
#include <X11/Xaw3d/SmeLine.h>

#else

//#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>

#endif

extern "C" {
#include <X11/At/Plotter.h>	// Athena Plotter Widget Set V6.0
#include <X11/At/XYAxis.h>	// with minor patches
#define GNANS_HACK
#include <X11/At/XYLinePlot.h>
}

#define widgetlog if (do_logging && trace_widgets) va_log_event

const int number_of_marker_types = 9;
const int number_of_line_styles = 12;
const int number_of_colors = 8;
const double ZOOM = 100000;	// sets max magnification (was 100)
const char * const linTicFormat = "%g";	// format like 'sprintf'
const char * const logTicFormat = "%g";	// both labeling

// this are the named colors
static const char * const coltab[number_of_colors] = {
  "red",
  "yellow",
  "deep sky blue",
  "white",
  "orange",
  "green",
  "magenta",
  "cyan"
};

Plotter *Plotter::head_plotter_list = NULL;
bool Plotter::use_color = false;

// General utility function
static unsigned long
AllocColorByName(const Widget w, const char *name)
{
  Display *d = XtDisplay(w);
  int s = DefaultScreen(d);
  Colormap col_map = DefaultColormap(d, s);
  XColor col, no_effect;
  if (XAllocNamedColor(d, col_map, name, &col, &no_effect))
    return col.pixel;
  else {
    gnans_error(non_fatal,
		"Couldn't allocate color named `%s'. Using black instead.",
		name);
    return BlackPixel(d, s);
  }
}

// Callbacks. These are just wrappers since member funcs cannot be passed
// as callbacks. (Xt and C, sigh...)
static void
PlotType(Widget /*mw*/, XtPointer client_data, XtPointer /*cd*/)
{
  widgetlog("PlotType");
  ((Plotter *) client_data)->plot_type();
}

static void
Zooming(Widget /*w*/, XtPointer client_data, AtRectangleCallbackData *cad)
{
  widgetlog("Zooming");
  ((Plotter *) client_data)->zooming(cad);
}

static void
Zoom(Widget /*w*/, XtPointer client_data, XtPointer /*cad*/) {
  widgetlog("zoom");
  ((Plotter *) client_data)->zoom();
}

static void
Grid(Widget /*w*/, XtPointer client_data, XtPointer /*cd*/)
{
  widgetlog("Grid");
  ((Plotter *) client_data)->grid();
}

static void
Subgrid(Widget /*w*/, XtPointer client_data, XtPointer /*cd*/)
{
  widgetlog("Subgrid");
  ((Plotter *) client_data)->subgrid();
}

static void 
PrintPostscript(Widget /*w*/, XtPointer client_data, XtPointer /*cd*/)
{
  widgetlog("PrintPostscript");
  ((Plotter *) client_data)->print_postscript();
}

static void
Linlog_select(Widget /*w*/, XtPointer client_data, XtPointer /*call_data*/)
{
  PlotterIntCallback *l = (PlotterIntCallback *) client_data;
  widgetlog("Linlog_select: %d", l->n);
  l->plotter_control->linlog_select(Plotter::LinLogType(l->n));
}

static void 
DeletePlotter(Widget /*w*/, XtPointer client_data, XtPointer /*cd*/)
{
  widgetlog("DeletePlotter");
  delete (Plotter *) client_data;
}

Plotter::Plotter(const Widget parent, const PlotterControl *creator)
{
  const PlotterControl *controller = creator;
  unsigned int independent_idx;

  // First find the independent variable
  bool found = false; // Just to check that exactly one variable is indep.
  for (unsigned int i = 0; i < number_of_stores; i++) {
    if (controller->vars()[i].x_selected()) {
      if (found)
	gnans_error(impossible, "More than one independent variable found");
      else {
	independent_idx = i;
	found = true;
      }
    }
  }
  if (!found)
    gnans_error(impossible, "No independent variable found");
  
  unsigned int *deps = new unsigned int[number_of_stores];
  unsigned int j = 0;
  for (unsigned int  i = 0;  i < number_of_stores; i++)
    if (controller->vars()[i].y_selected())
      deps[j++] = i;
  unsigned int no_deps = j;

  init(parent, independent_idx, no_deps, deps);
  delete [] deps;
}

Plotter::Plotter(const Widget parent, unsigned int independent_idx,
		 unsigned int no_deps, unsigned int *deps) 
{
  init(parent, independent_idx, no_deps, deps);
}

static int
findit(char **str)
{
  char *s = strchr(*str, ',');
  if (s)
    *s = '\0';
  int idx = -1;

  for (unsigned int i = 0; i < number_of_stores; i++)
    if (strcmp(stores[i].name.c_str(), *str) == 0)
      idx = i;

  if (idx == -1)
    printf("did not find %s", *str);
  else
    *str = s ? s + 1 : 0;
  return idx;
}     

Plotter::Plotter(const Widget parent, const char *var_list)
{
  if (!toplevel) {
    gnans_error(non_fatal, "X not initialized, cannot plot");
    return;
  }
  char *str = new char[MAXSTRINGSIZE];
  strcpy(str, var_list);
  unsigned int deps[number_of_stores];
  unsigned int var_index = 0;
  int indep_index = findit(&str);
  if (indep_index == -1) {
    gnans_error(non_fatal, "Independent variable %s not found", str);
    return;
  }
  if (!str) {
    gnans_error(non_fatal, "No dependent variable given");
    return;
  }
  for (var_index = 0; str; var_index++) {
    deps[var_index] = findit(&str);
    if (deps[var_index] == -1) {
      gnans_error(non_fatal, "Dependent variable %s not found", str);
      return;
    }
  }
  
  if (do_logging)
    va_log_event("Plotting %s", var_list);
  init(parent, indep_index, var_index, deps);
}

void
Plotter::init(const Widget parent, unsigned int independent_idx,
	       unsigned int no_deps, unsigned int *deps) 
{
  Widget button, menu, entry;
  
  if (independent_idx > number_of_stores - 1|| no_deps > number_of_stores) {
    gnans_error(non_fatal, "Cannot plot");
    return;
  }
  independent_index = independent_idx;
  no_dependents = no_deps;
  dependents = new unsigned int[no_deps];
  for (unsigned int i = 0; i < no_dependents; i++) {
    if (deps[i] > number_of_stores - 1) {
      gnans_error(non_fatal, "Cannot plot");
      return;
    }
    dependents[i] = deps[i];
  }

  use_color = XDisplayCells(XtDisplay(parent),
			    DefaultScreen(XtDisplay(parent))) > 2;

  linlog_table[LinLin].name = "Lin-Lin";
  linlog_table[LinLog].name = "Lin-Log";
  linlog_table[LogLin].name = "Log-Lin";
  linlog_table[LogLog].name = "Log-Log";
  
  linlog_table[LinLin].selected = true;
  current_axes_type = LinLin;

  point_p = current_system.system_type == discrete;
  zooming_mode = false;
  grid_p = true;
  subgrid_p = false;

  plot = new Widget[number_of_stores];
  for (unsigned int i = 0; i < number_of_stores; i++)
    plot[i] = 0;

  // window
  plotShell = XtVaCreatePopupShell("plotShell", transientShellWidgetClass,
				   parent, XtNtitle,
				   make_popup_title(title, "Gnans Plot"),
				   NULL);
  // a reserved form within the window
  form = XtVaCreateManagedWidget("form", formWidgetClass, plotShell, NULL);

  // now the buttons of the plot-shell

  // Lin/Log-Pull-Down-Menu (exact one selection)
  llbutton = XtVaCreateManagedWidget("linlog_menu_button",
				     menuButtonWidgetClass,
				     form, XtNleft, XtChainLeft, NULL);
  menu = XtVaCreatePopupShell("menu", simpleMenuWidgetClass, llbutton, NULL);
  entry = XtVaCreateManagedWidget("line", smeLineObjectClass, menu, NULL);
  mark = XCreateBitmapFromData(XtDisplay(toplevel),
			       RootWindowOfScreen(XtScreen(toplevel)),
			       (const char *) xlogo16_bits,
			       xlogo16_width, xlogo16_height);
  
  for (unsigned int i = 0; i < 4; i++) {
    entry = XtVaCreateManagedWidget(linlog_table[i].name,
				    smeBSBObjectClass,  menu, XtNleftBitmap,
				    linlog_table[i].selected ? mark : None,
				    NULL);
    linlog_table[i].widget = entry;
    linlog_table[i].callback.plotter_control = this;
    linlog_table[i].callback.n = i;
    XtAddCallback(entry, XtNcallback, Linlog_select,
		  &linlog_table[i].callback);
  }

  typebutton = XtVaCreateManagedWidget("plotType", commandWidgetClass, form,
				       XtNfromHoriz, llbutton, NULL);
  XtAddCallback(typebutton, XtNcallback, PlotType, this);
  zoom_button = XtVaCreateManagedWidget("zoom", commandWidgetClass, form,
					XtNfromHoriz, typebutton, NULL);
  XtAddCallback(zoom_button, XtNcallback, Zoom, this);
  gridbutt = XtVaCreateManagedWidget("grid", commandWidgetClass, form,
				     XtNfromHoriz, zoom_button, NULL);
  XtAddCallback(gridbutt, XtNcallback, Grid, this);
  subgridbutt = XtVaCreateManagedWidget("subgrid", commandWidgetClass, form,
					XtNfromHoriz, gridbutt, NULL);
  XtAddCallback(subgridbutt, XtNcallback, Subgrid, this);
  print_button = XtVaCreateManagedWidget("print", commandWidgetClass, form,  
					 XtNfromHoriz, subgridbutt, NULL);
  XtAddCallback(print_button, XtNcallback, PrintPostscript, this);
  button = XtVaCreateManagedWidget("ok", commandWidgetClass, form,
				   XtNfromHoriz, print_button, NULL);
  XtAddCallback(button, XtNcallback, DeletePlotter, this);

  // create plotter within the form 
  plotter = XtVaCreateManagedWidget("plotter", atPlotterWidgetClass, form,
				    //XtNtitle,
				    //(char *)current_system.system_name,
				    XtNshowLegend, True,
				    NULL);
  
  // create X axis 
  xaxis = XtVaCreateWidget("xaxis", atXYAxisWidgetClass, plotter,
			   XtNlabel, "Independent", // Don't delete this line!
			   XtNdrawGrid, True, 
			   XtNdrawSubgrid, False,
			   XtNaxisWidth, 1,
			   XtNlinTicFormat, linTicFormat,
			   XtNlogTicFormat, logTicFormat,
			   NULL); 
  
  // create Y axis
  yaxis = XtVaCreateWidget("yaxis", atXYAxisWidgetClass, plotter,
			   XtNlabel, "",  // Don't delete this line!
			   XtNaxisWidth, 1,
			   XtNdrawGrid, True,
			   XtNdrawSubgrid, False,
			   XtNvertical, True,
			   XtNlinTicFormat, linTicFormat,
			   XtNlogTicFormat, logTicFormat,
			   NULL);
  
  // attach x and y to the plotter 
  XtVaSetValues(plotter, XtNxAxis, xaxis, XtNyAxis, yaxis, NULL);
  XtVaSetValues(plotter, XtNleft, XtChainLeft, XtNfromVert, button, NULL);
  
  // Now attaching the data to the plotter
  attach_data(plotter);
  plot_type(false, false);

  subgrid(false, false);
  grid(false, false);

  // popup the window with the plot
  popup_position(plotShell, parent, XtGrabNone, false);

  // set label for zoom-button
  set_grid_label();
  set_subgrid_label();
  set_zoom_label();
  set_plot_type_label();

  insert_in_list();
}

Plotter::~Plotter()
{
  XtDestroyWidget(plotShell);
		  
  delete [] plot;

  // Remove me from the linked list
  remove_me(head_plotter_list);
}

void
Plotter::remove_me(Plotter * &p)
{
  // Don't make trouble if not found
  if (p) {
    if (p == this)
      // Found it, now remove
      p = next;
    else
      remove_me(p->next);
  }
}

// does the selected plot mode if possible and sets the mark
void
Plotter::linlog_select(LinLogType n)
{  
  bool xerr = (n == LogLin || n == LogLog) && !x_all_positive();
  bool yerr = (n == LinLog || n == LogLog) && !y_all_positive();

  if (xerr && yerr) {
    gnans_error(non_fatal,
		"Log-Log plot not possible since nonpositive\n"
		"values found in both horizontal and vertical data.");
    return;
  }
  if (xerr) {
    gnans_error(non_fatal,
                "Logarithmic horizontal axis not possible since nonpositive\n"
                "values found in horizontal data.");
    return;
  }
  if (yerr) {
    gnans_error(non_fatal,
                "Logarithmic vertical axis not possible since nonpositive\n"
                "values found in vertical data.");
    return;
  }

  XtVaSetValues(xaxis, XtNaxisTransform,
		(n == LogLin || n == LogLog)
		? AtTransformLOGARITHMIC : AtTransformLINEAR,
		NULL);
  
  XtVaSetValues(yaxis, XtNaxisTransform,
		(n == LinLog || n == LogLog)
		? AtTransformLOGARITHMIC : AtTransformLINEAR,
		NULL);

  // set mark at the right position
  for (int i = int(LinLin); i <= int(LogLog); i++) {
    linlog_table[i].selected = i == n;
    XtVaSetValues(linlog_table[i].widget,		
		  XtNleftBitmap, linlog_table[i].selected ? mark : None,
		  NULL);
  }

  current_axes_type = n;
}

// Attach the data to the plotter
void
Plotter::attach_data(const Widget parent)
{
  unsigned int color_index = 0;
  unsigned int number_of_plots = 0;

  std::string independent_label(stores[independent_index].name);
  std::string dependent_label = "";

  for (unsigned int i = 0; i < no_dependents; i++) {
    if (use_color) {
      // create this plot and set color for it
      plot[i] = XtVaCreateWidget("plot", atXYLinePlotWidgetClass, parent,
				 XtNforeground,
				 AllocColorByName(parent, coltab[color_index]),
				 XtNlegendName,
				 stores[dependents[i]].name.c_str(),
				 NULL);
    } else {
      // create this plot without color but with different styles
      // the foreground is white, the plot color is black
      // (unless overriden in app-defaults).
      plot[i] = XtVaCreateWidget("plot", atXYLinePlotWidgetClass, parent,
				 XtNforeground, "white",
				 XtNlegendName,
				 stores[dependents[i]].name.c_str(),
				 XtNplotLineStyle, color_index, NULL);
    }
	
    // Now attaching the data to the plotter.
    // Main-function of the plotter.
    // Every plot has his own widget, the data is stored in an array form
    // typed AtDouble, the size of the fields is double, beginning point
    // is first (0) and sample_no points should be plot
    AtXYLinePlotAttachData(plot[i],
			   // x-axis
			   (XtPointer) stores[independent_index].trajectory,
			   AtDouble, sizeof(double), 

			   // y-axis
			   (XtPointer) stores[dependents[i]].trajectory,
			   AtDouble, sizeof(double),
			   0, sample_no);
    AtXYPlotAddDiscontinuities((AtXYPlotWidget) plot[i],
			       Store::get_no_of_trajectories(),
			       Store::get_discontinuities());
    number_of_plots++;
    color_index++;
    color_index %= use_color ? number_of_colors : number_of_line_styles;

    if (dependent_label != "")
      dependent_label += ", ";
    dependent_label += stores[dependents[i]].name;

  }
  one_plot = number_of_plots == 1;

  // set labeling for the axis
  //XtVaSetValues(yaxis, XtNlabel, (char *) dependent_label, NULL);
  XtVaSetValues(xaxis, XtNlabel, independent_label.c_str(), NULL);
}

// Just a utility func
static inline
double max(double x, double y)
{
  return x > y ? x : y;
}

static inline
unsigned int max(unsigned int x, unsigned int y)
{
  return x > y ? x : y;
}

static inline
double min(double x, double y)
{
  return x < y ? x : y;
}

static inline
double clip(double x, double x_min, double x_max)
{
  return max(min(x, x_max), x_min);
}

// Compute whether logarithmic plots are possible.
bool
Plotter::x_all_positive()
{
  double x_min, x_max;
  XtVaGetValues(xaxis, XtNmin, &x_min, XtNmax, &x_max, NULL);
  return x_all_positive(x_min, x_max);
}

bool
Plotter::x_all_positive(double x_min, double x_max)
{
  for (unsigned int i = 0; i < sample_no; i++)
    if (clip(stores[independent_index].trajectory[i], x_min, x_max) <= 0.0)
      return false;
  return true;
}

bool
Plotter::y_all_positive()
{
  double y_min, y_max;
  XtVaGetValues(yaxis, XtNmin, &y_min, XtNmax, &y_max, NULL);
  return y_all_positive(y_min, y_max);
}

bool
Plotter::y_all_positive(double y_min, double y_max)
{
  for (unsigned int var = 0; var < no_dependents; var++)
    for (unsigned int i = 0; i < sample_no; i++)
      if (clip(stores[dependents[var]].trajectory[i], y_min, y_max) <= 0.0)
	return false;
  
  return true;
}

// Update button label
void
Plotter::set_plot_type_label()
{
  XtVaSetValues(typebutton, XtNlabel, point_p ? "Line Plot" : "Point Plot",
		NULL);
}

// Toggles between line plot and point plot
// Different styles or marks whether using color or not
void
Plotter::plot_type(bool update_labels /*= true*/, bool toggle /*= true*/)
{
  if (toggle)
    point_p = !point_p; 

  unsigned int index = 0;	// Index of color OR marker type
  unsigned int i;
  for (i = 0; i < no_dependents; i++) {
    if (plot[i]) {
      if (point_p) {
	// user wishes point plot
	XtVaSetValues(plot[i],
		      XtNplotLineStyle, AtLineINVALID, // default style
		      XtNplotLineType, AtPlotPOINTS,
		      XtNplotMarkType,
		      (use_color || one_plot
		       ? AtMarkPIXEL : (AtPlotMarkType) index),
		      NULL);
      } else {
	// user wishes line plot
	XtVaSetValues(plot[i],
		      XtNplotLineStyle, use_color ? 0 : index,
		      XtNplotLineType, AtPlotLINES,
		      NULL);
      }
     
    }
    index++;
    index %= point_p ? number_of_marker_types : number_of_line_styles;
  }

  if (update_labels)
    set_plot_type_label();
}

// callback func of zoom()
// sets new bounds for the zoomed plot
void
Plotter::zooming(const AtRectangleCallbackData *cad)
{
  double x1 = cad->x11;
  double x2 = cad->x12;
  double xl = fabs(x2-x1);
  double y1 = cad->y11;
  double y2 = cad->y12;
  double yl = fabs(y2-y1);
  
 
  // First check that the user is not requesting too large magnification,
  // in that case, just barf.
  if (xl < minlenx && yl < minleny) {
    gnans_error(non_fatal,
		"Too large magnification (>%g times) requested in both\n"
		"horizontal and vertical direction.", ZOOM);
    return;
  }
  if (xl < minlenx) {
    gnans_error(non_fatal,
		"Too large magnification (>%g times) requested in "
		"horizontal direction.",
		ZOOM);
    return;
  }
  if (yl < minleny) {
    gnans_error(non_fatal,
		"Too large magnification (>%g times) requested in "
		"vertical direction.",
		ZOOM);
    return;
  }

  XtVaSetValues(xaxis,
		XtNmin, &x1,
		XtNmax, &x2,
		XtNautoScale, False,
		NULL);
  XtVaSetValues(yaxis,
		XtNmin, &y1,
		XtNmax, &y2,
		XtNautoScale, False,
		NULL);
}

void
Plotter::zoom() {
  zooming_mode = !zooming_mode;

  if (zooming_mode) {
    XtVaGetValues(xaxis, XtNmin, &oldxmin, XtNmax, &oldxmax, NULL); // save old
    XtVaGetValues(yaxis, XtNmin, &oldymin, XtNmax, &oldymax, NULL);  
    minlenx = (oldxmax-oldxmin+.000001) / ZOOM; // compute max zoom rate
    minleny = (oldymax-oldymin+.000001) / ZOOM;
    XtAddCallback(plotter, XtNdragCallback, (XtCallbackProc) &Zooming, this); // zoom
  } else {
    // reset the axes to their original size
    
    // First check if we have to turn logarithmic axes to linear ones
    if (
	((current_axes_type == LogLin || current_axes_type == LogLog)
	 && !x_all_positive(oldxmin, oldxmax))
	|| ((current_axes_type == LinLog || current_axes_type == LogLog)
	    && !y_all_positive(oldymin, oldymax)))
      {
      gnans_error(message, "Resetting to Lin-Lin axes");
      linlog_select(LinLin);
    }
          
    XtVaSetValues(xaxis,
		  XtNmin, &oldxmin,
		  XtNmax, &oldxmax,
		  XtNautoScale, True,
		  NULL);
    XtVaSetValues(yaxis,
		  XtNmin, &oldymin,
		  XtNmax, &oldymax,
		  XtNautoScale, True,
		  NULL);
    XtRemoveAllCallbacks(plotter, XtNdragCallback);
  }
  // Change button label
  set_zoom_label();
}

void
Plotter::set_zoom_label()
{
  XtVaSetValues(zoom_button, XtNlabel, zooming_mode ? "Unzoom" : "Zoom", NULL);
}

// Toggles grid
void
Plotter::grid(bool update_label /*= true*/, bool toggle /*= true*/)
{
  if (toggle)
    grid_p = !grid_p;

  XtVaSetValues(xaxis, XtNdrawGrid, int(grid_p), XtNdrawOrigin, int(grid_p),
		NULL);
  XtVaSetValues(yaxis, XtNdrawGrid, int(grid_p), XtNdrawOrigin, int(grid_p),
		NULL);

  // Disable subgrid if no grid
  XtVaSetValues(subgridbutt, XtNsensitive, int(grid_p), NULL);
  
  // Change button label
  if (update_label)
    set_grid_label();
}

void
Plotter::set_grid_label()
{
  XtVaSetValues(gridbutt, XtNlabel, grid_p ? "No Grid" : "Grid", NULL); 
}

// Toggles subgrid
void
Plotter::subgrid(bool update_label /*= true*/, bool toggle /*= true*/)
{
  if (toggle)
    subgrid_p = !subgrid_p;

  XtVaSetValues(xaxis, XtNdrawSubgrid, int(subgrid_p), NULL);
  XtVaSetValues(yaxis, XtNdrawSubgrid, int(subgrid_p), NULL);

  // Change button label
  if (update_label)
    set_subgrid_label();
}

void
Plotter::set_subgrid_label()
{
  XtVaSetValues(subgridbutt, XtNlabel, subgrid_p ? "No Subg." : "Subgrid",
		NULL);
}

// Generates a plot with filename
//   system-name + ".gnans." + PID + "." + actual number of the session + ".ps"
void
Plotter::print_postscript()
{
  char filenam[MAXSTRINGSIZE];
  static unsigned int psnr = 1;
  sprintf(filenam, "%s.gnans.%d.%d.ps",
	  current_system.system_name.c_str(), int(getpid()), psnr++);

  // An awful kludge to get different line styles if the user
  // has a color display
  bool color_saved = use_color;
  if (use_color) {
    use_color = false;
    plot_type(false, false);
  }
  
  // This function dumps the plot to a file in the given size and portrait mode
  AtPlotterGeneratePostscript(filenam, (AtPlotterWidget) plotter,
			      "Gnans Plot", 10, 300, 580, 680, FALSE);

  if (color_saved) {
    use_color = true;
    plot_type(false, false);
  }
  gnans_error(message, "Postscript file ``%s'' created.", filenam);
}

void
Plotter::insert_in_list()
{
  next = head_plotter_list; 
  head_plotter_list = this;
}

// Disable everything except for the done button,
// and leave it hanging there without any references kept.
void
Plotter::disable()
{
  // disable plot from actual trajectories
  XtVaSetValues(plotter,XtNautoRedisplay, False, NULL);
  XtRemoveAllCallbacks(plotter, XtNlayoutCallback); // Doesn't work?!
  for (unsigned int i = 0; i < number_of_stores; i++)
    if (plot[i])
      AtXYLinePlotAttachData(plot[i], 
			     NULL, AtDouble, sizeof(double), 
			     NULL, AtDouble, sizeof(double),
			     0, 0);
     
  // disable all buttons from plot-form, except the done_button
  XtVaSetValues(llbutton, XtNsensitive, False, NULL);
  XtVaSetValues(typebutton, XtNsensitive, False, NULL);
  XtVaSetValues(zoom_button, XtNsensitive, False, NULL);
  XtVaSetValues(gridbutt, XtNsensitive, False, NULL);
  XtVaSetValues(subgridbutt, XtNsensitive, False, NULL);
  XtVaSetValues(print_button, XtNsensitive, False, NULL);
}

void
Plotter::disable_plotters(Plotter *p)
{
  if (p) {
    disable_plotters(p->next);
    p->disable();
  }
}

void
Plotter::disable_plotters()
{
  disable_plotters(head_plotter_list);
  head_plotter_list = NULL;
} 
