/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

// $Id: PlotterControl.cc,v 1.6 1994/08/19 18:38:31 bengt Exp $

// Based on a file by Patrick Rosenberg 

#ifdef __GNUG__
#pragma implementation
#endif

#include "externals.h"
#ifdef HAVE_IEEEFP_H
#include <ieeefp.h>		// For finite()
#endif

static void Select_Dependent(Widget, XtPointer, XtPointer);
static void Select_Independent(Widget, XtPointer, XtPointer);
static void DeletePlotterControl(Widget, XtPointer, XtPointer);
static void CreatePlotter(Widget, XtPointer, XtPointer);

#include "PlotterControl.h"
typedef char *String;
#include "Plotter.h"

#include <X11/Intrinsic.h>
#include <X11/bitmaps/xlogo16>

#ifdef USEXAW3DINCLUDES

#include <X11/Xaw3d/Box.h>
#include <X11/Xaw3d/MenuButton.h>
#include <X11/Xaw3d/SimpleMenu.h>
#include <X11/Xaw3d/SmeBSB.h>
#include <X11/Xaw3d/SmeLine.h>

#else

#include <X11/Xaw/Box.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>

#endif

#define widgetlog if (do_logging && trace_widgets) va_log_event

// Static variables in this file:
PlotterControl *PlotterControl::head_plotter_control_list = NULL;

// Callbacks. These are just wrappers since member funcs cannot be passed
// as callbacks. (Xt and C, sigh...)

static void
CreatePlotter(Widget parent, XtPointer client_data, XtPointer /*cd*/)
{
  widgetlog("Creating Plotter");
  new Plotter(parent, (PlotterControl *) client_data);
}

static void 
Select_Dependent(Widget /*w*/, XtPointer s, XtPointer /*call_data*/)
{
  PlotterControlIntCallback *t = (PlotterControlIntCallback *) s;
  widgetlog("Select_Dependent: %d", t->n);
  t->plotter_control->select_dependent(t->n);
}

static void 
Select_Independent(Widget /*w*/, XtPointer client_data, XtPointer /*call_data*/)
{
  PlotterControlIntCallback *m = (PlotterControlIntCallback *) client_data;
  widgetlog("Select_Independent: %d", m->n);
  m->plotter_control->select_independent(m->n);
}

static void 
DeletePlotterControl(Widget /*w*/, XtPointer client_data, XtPointer /*cd*/)
{
  widgetlog("DeletePlotterControl");
  delete (PlotterControl *) client_data;
}

PlotCandidate::PlotCandidate()
{
  x_wdgt = y_wdgt = NULL;
  x_sel = y_sel = false;
  nam = "";
}

PlotterControl::PlotterControl(const Widget mother_widget)
{
  if (! X) {
    gnans_error(impossible, "Plot routines only supported under X");
    //this = NULL;
    return;
  }

  // Check that there are no Infinity or NaNs in the data
  // (It is slightly overkill to check all variables, but faster to program :-)
  unsigned int i;
  for (i = 0; i < number_of_stores; i++) {
    unsigned int j;
    for (j = 0; j < sample_no; j++)
      if (! finite(stores[i].trajectory[j])) {
	gnans_error(non_fatal, "Infinity or NaN found in data. Not plotting.");
	return;
      }
  }

  // First allocate space for tables
  variables = new PlotCandidate[number_of_stores];

  // Create bitmap for marking selected items
  mark = XCreateBitmapFromData(XtDisplay(toplevel),
                               RootWindowOfScreen(XtScreen(toplevel)),
                               (const char *) xlogo16_bits,
                               xlogo16_width, xlogo16_height);
  
  // Create Popup
  pop = XtVaCreatePopupShell("popup", transientShellWidgetClass, mother_widget,
			     XtNtitle, make_popup_title(title, "Plot"), NULL);
  Widget box = XtVaCreateManagedWidget("box", boxWidgetClass, pop, NULL);
  
  // Initialize the structures of the Vertical- and Horizontal-Pull-Down-Menus
  for (i = 0; i < number_of_stores; i++)
    variables[i].setname(stores[i].name);

  // default setting: first values automatically selected
  variables[0].x_sel = true;
  variables[number_of_stores > 1 ? 1 : 0].y_sel = true;
      
  Widget button = XtVaCreateManagedWidget("dep_menu_button",
					  menuButtonWidgetClass,
					  box, NULL);
  Widget menu = XtVaCreatePopupShell("menu", simpleMenuWidgetClass, button,
				     NULL);
  Widget entry = XtVaCreateManagedWidget("line", smeLineObjectClass, menu,
					 NULL);  
  for (i = 0; i < number_of_stores; i++) {
    entry = XtVaCreateManagedWidget(variables[i].name(), smeBSBObjectClass,
				    menu, XtNleftBitmap,
				    variables[i].y_selected() ? mark : None,
				    NULL);
    variables[i].y_wdgt = entry;
    variables[i].callback.plotter_control = this;
    variables[i].callback.n = i;
    XtAddCallback(entry, XtNcallback, Select_Dependent,
		  &variables[i].callback);
  }
    
  // Horizontal-Menu
  button = XtVaCreateManagedWidget("indep_menu_button", menuButtonWidgetClass,
				   box, NULL);
  menu = XtVaCreatePopupShell("menu", simpleMenuWidgetClass, button, NULL);
  entry = XtVaCreateManagedWidget("line", smeLineObjectClass, menu, NULL);
  for (i = 0; i < number_of_stores; i++) {
    entry = XtVaCreateManagedWidget(variables[i].name(), smeBSBObjectClass,
				    menu, XtNleftBitmap,
				    variables[i].x_selected() ? mark : None,
				    NULL); 
    variables[i].x_wdgt = entry;
    variables[i].callback.plotter_control = this;
    variables[i].callback.n = i;
    XtAddCallback(entry, XtNcallback, Select_Independent,
		  &variables[i].callback);
  }

  // Plot-button
  plot_button = XtVaCreateManagedWidget("plotting", commandWidgetClass, box,
					NULL);
  XtAddCallback(plot_button, XtNcallback, CreatePlotter, this);
  
  // Done-button
  button = XtVaCreateManagedWidget("ok", commandWidgetClass, box, NULL);
  XtAddCallback(button, XtNcallback, DeletePlotterControl, this);
  
  // Ok, do popup
  popup_position(pop, mother_widget, XtGrabNone, false);

  // Put it first it in the list
  insert_in_list();

}

PlotterControl::~PlotterControl()
{
  // Let my children plotters survive me, so popdown, not destroy
  XtPopdown(pop);
  
  delete [] variables;

  // Remove me from the linked list
  remove_me(head_plotter_control_list);
}

void
PlotterControl::remove_me(PlotterControl * &p)
{
  if (p == this)
    // Found it, now remove
    p = next;
  else
    remove_me(p->next);  
}

// for selecting the dependent variables and setting the mark
void
PlotterControl::select_dependent(const unsigned int n)
{
  variables[n].y_select();

  // Check that at least one dependent variable is selected,
  // otherwise re-set this one
  bool ok = false;
  for (unsigned int i = 0; i < number_of_stores; i++)
    ok = ok || variables[i].y_selected();
  
  if (!ok)
    variables[n].y_select(true);
  
  XtVaSetValues(variables[n].y_widget(),
		XtNleftBitmap, variables[n].y_selected() ? mark : None,
		NULL); 
}

// for selecting the independent variable and setting the mark
void
PlotterControl::select_independent(const unsigned int n)
{
  for (unsigned int i = 0; i < number_of_stores; i++) {
    variables[i].x_select(i == n);
    XtVaSetValues(variables[i].x_widget(),
		  XtNleftBitmap, variables[i].x_selected() ? mark : None,
		  NULL);
  }
}

/* static */ void
PlotterControl::destroy_plotter_controls(PlotterControl *p)
{
  if (p) {
    destroy_plotter_controls(p->next);
    delete p;
  }
}

/* static */ void
PlotterControl::destroy_plotter_controls()
{
  destroy_plotter_controls(head_plotter_control_list);
}

void
PlotterControl::insert_in_list()
{
  next = head_plotter_control_list; 
  head_plotter_control_list = this;
}

/* static */ void
PlotterControl::enable_plotter_controls(PlotterControl *p, bool enable)
{
  if (p) {
    enable_plotter_controls(p->next, enable);
    XtVaSetValues(p->plot_button, XtNsensitive, int(enable), NULL);
  }
}

/* static */ void
PlotterControl::enable_plotter_controls(bool enable)
{
  enable_plotter_controls(head_plotter_control_list, enable);
}
