%{
/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

/* $Id: cmdparse.y,v 1.10 1994/08/24 09:58:29 bengt Exp $ */

/*
 * This file defines the grammar for the script language.
 */

/*
 * yyparse should return:
 * 0 if everything has been successfully completed,
 * 1 on nonrecoverable error,
 * 2 if interrupted by the clock
 */

#define YYDEBUG 1

#ifdef __GNUC__
//#include <builtin.h>		/* pow(long, long) */
#endif

#include <math.h>
#include <stdlib.h>
#include "externals.h"
#include "utillib.h"
typedef char *String;           // ??????????????? FIXME
#include "Plotter.h"

extern "C" char *tilde_expand(char *); /* From Gnu readline */
#define PROMPT do_prompt(); \
                if (abort_script) return 1;\
		if (has_been_alarmed) return 2;
#define SCRIPT_LOG if (do_logging) va_log_event

#define HELPTEXT \
  "Gnans is documented in the Texinfo format, which allows the same file\n"\
  "to generate both an online, hypertext help system, and a version which\n"\
  "can be printed with TeX.  Please consult this file, either reading it\n"\
  "online with, e.g. Gnu Emacs or another info reader, or read the\n"\
  "printed documentation.\n"
/*
 * Assign a value to a variable, provided that the variable is not readonly.
 * Otherwise barf.
 */
static double
maybe_assign(symrec *variable, double value)
{
  if (! variable->readonly) {
    *variable = value;
    SCRIPT_LOG("%s = %g", variable->name, value);
  } else
    printf("\a`%s' is readonly!\n", variable->name);
  return value;
}

static void
assign_ans(double x)
{
  *(getsym("ans")) = x;
}

/*
 * Some junk to get rid of warnings on labels not used.
 */
#ifdef __GNUC__
#ifdef YYBISON
#define SHUT_UP_GCC_WARNINGS \
  if (0) \
    goto yyerrlab1; \
  else \
    goto yybackup;
#else
#define SHUT_UP_GCC_WARNINGS \
  if (0) \
    goto yyerrlab; \
  else \
    goto yynewstate;
#define free(p) c_free((generic_ptr) p)
#endif
#else
#define SHUT_UP_GCC_WARNINGS
#endif

%}

/* Data type returned from yylex */
%union {
  int action;			/* For returning gnans commands */
  char *text;			/* For returning strings */
  double val;			/* For returning numbers */
  symrec *tptr;			/* For returning symbol-table pointers */
}

%token <val>    NUM		/* Simple double precision number  */
%token <tptr>   VAR FNCT	/* Variable and Function */
%token <text>	STRING		/* Doublequote enclosed strings */
%type  <val>    exp
%type  <val>    stmt
%type  <action> display_unit
%type  <action>	axis_type
%type  <text>   plot_vars
%type  <text>   pe_plot_vars
%type  <action> plot_options
%type  <action> plot_option
%type  <action> boolean
%token <action> START
%token <action> QUIT
%token <action> HELP
%token <action> DISPLAY
%token <action> SET
%token <action> SWITCH
%token <action> SAVE_STATE
%token <action> RESTORE_STATE
%token <action> ON
%token <action> OFF
%token <action> PLOT
%token <action> AXIS
%token <action> LINEPLOT
%token <action> LINLIN
%token <action> LINLOG
%token <action> LOGLIN
%token <action> LOGLOG
%token <action> GNANS_TRUE
%token <action> GNANS_FALSE
%token <action> FUNCTIONS
%token <action> INTERNALS
%token <action> RESOURCES
%token <action> PARAMETERS
%token <action> RANDOMS
%token <action> STATES
%token <action> TIMES
%token <action> SYSTEM
%token <action> VARIABLES
%token <action> STORE
%token <action> UNSTORE
%token <action> CONTINUE
%token <action> CONTINUE_INIT
%token <action> ALGORITHM
%token <action> FORMAT

%right '=' PLUSE MINUSE TIMESE SLASHE PERCENTE
%right '?' ':'
%left OR
%left AND
%left '|'
%left '&'
%left EQ NE
%left '<' '>' GE LE
%left LL GG
%left '-' '+'
%left '*' '/' '%'
%left UMINUS UPLUS '!' '~' PLUSPLUS MINUSMINUS
%right '^'			/* Exponentiation */

%%
input:    /* empty */
	| input stmt
;

stmt:
	  '\n'		{ PROMPT; }
	| ';'		{}
	| exp ';'  	{ assign_ans($1); }
	| exp '\n'  	{
                          printf(floating_point_format.c_str(), $1);
                          assign_ans($1);
                          PROMPT;
                        }
	| cmd '\n' 	{}
	| cmd ';'  	{}
	| error '\n' 	{ yyerrok; PROMPT; }
;

cmd:      START 	{ integrate(init_all); PROMPT; }
	| CONTINUE	{ integrate(init_none); PROMPT; }
	| CONTINUE_INIT	{ integrate(init_states); PROMPT; }
	| HELP		{ printf(HELPTEXT); PROMPT; }
	| QUIT 		{ SCRIPT_LOG("Quit"); die("", 0);
			  SHUT_UP_GCC_WARNINGS; }
	| SYSTEM STRING	{ 
#ifdef DYNAMIC_LOADING
  char *name = tilde_expand($2);
  setup_system_dynamic(String(name));
  free(name);
  delete [] $2;
  update_symbol_table();
#else
  gnans_error(non_fatal,
	      "Sorry, this version of Gnans doesn't support dynamic loading");
  delete [] $2;
#endif
			  PROMPT; }
	| DISPLAY display_unit 
			{ Display_vars($2); PROMPT; }
	| SET ALGORITHM '=' STRING
			{ select_algorithm($4); delete [] $4; PROMPT; }
	| SET FORMAT '=' STRING
			{ select_format($4); delete [] $4; PROMPT; }
	| SWITCH STRING ON
			{ toggle_switch($2, true); delete [] $2; PROMPT; }
	| SWITCH STRING OFF
			{ toggle_switch($2, false); delete [] $2; PROMPT; }
	| STORE VAR 	{ store($2->name, true); PROMPT; }
	| UNSTORE VAR   { store($2->name, false); PROMPT; }
	| RESTORE_STATE '(' VAR ',' exp ')'
			{ restore_state($3->name, (unsigned int) $5); PROMPT; }
        | PLOT plot_options plot_vars 
                        { if (toplevel) {
                             Plotter *p = new Plotter(toplevel, $3);
                             if ($2 % 1000 == LINLOG)
                                p->linlog_select(Plotter::LinLog);
                             if ($2 % 1000 == LOGLIN)
                                p->linlog_select(Plotter::LogLin);
                             if ($2 % 1000 == LOGLOG)
                                p->linlog_select(Plotter::LogLog);
                             //if (($2 % 10000)/1000) {
                                //p->print_postscript();
                             //}
                             //if ($2/10000)
			     //   delete p;
                             if ($2/1000 == GNANS_TRUE) {
                                p->point_p=false;
                                p->plot_type(true, false);
                             }
                             if ($2/1000 == GNANS_FALSE) {
                                p->point_p=true;
                                p->plot_type(true, false);
                             }


                          } else
                             gnans_error(non_fatal, 
                                         "X not initialized, cannot plot");
                          delete [] $3;
                          PROMPT;
}
;

plot_options:
          /* empty */              { $$ = 0; }
        | plot_option plot_options { $$ = $1 + $2;}
;

plot_option:
          AXIS '=' axis_type     {$$ = $3;}
	| LINEPLOT '=' boolean   {$$ = 1000*$3;}
;

boolean:
	  GNANS_TRUE             { $$ = GNANS_TRUE; }
        | GNANS_FALSE            { $$ = GNANS_FALSE; }

axis_type:
          LINLIN { $$ = LINLIN; }
	| LINLOG { $$ = LINLOG; }
        | LOGLIN { $$ = LOGLIN; }
	| LOGLOG { $$ = LOGLOG; }
;

plot_vars:
	  VAR VAR pe_plot_vars { $$ = new char[MAXSTRINGSIZE];
                                 strcpy($$, $1->name);
                                 strcat($$, ",");
	                         strcat($$, $2->name);
				 if (strlen($3)) {
				    strcat($$, ",");
				    strcat($$, $3);
                                 }
                                 delete [] $3;
}
;

pe_plot_vars:
	  /* Empty */      { $$ = new char[MAXSTRINGSIZE];
                             $$[0] = '\0'; }
	| VAR pe_plot_vars { $$ = new char[MAXSTRINGSIZE];
                             strcpy($$, $1->name);
			     if (strlen($2)) {
				strcat($$, ",");
				strcat($$, $2);
                             }
                             delete [] $2;
}
;

display_unit:
  	  /* empty */ 	{ $$ = 0; }
        | STATES 	{ $$ = STATES; }
        | VARIABLES 	{ $$ = VARIABLES; }
        | PARAMETERS 	{ $$ = PARAMETERS; }
        | FUNCTIONS 	{ $$ = FUNCTIONS; }
        | TIMES 	{ $$ = TIMES; }
        | INTERNALS 	{ $$ = INTERNALS; }
        | RESOURCES 	{ $$ = RESOURCES; }
        | RANDOMS	{ $$ = RANDOMS; }
;

exp:      NUM           { $$ = $1; }
	| VAR           { $$ = *$1; }
	| VAR '=' exp   { $$ = maybe_assign($1, $3); }
	| VAR PLUSE exp { $$ = maybe_assign($1, double(*$1) + $3); }
	| VAR MINUSE exp { $$ = maybe_assign($1, double(*$1) - $3); }
	| VAR TIMESE exp { $$ = maybe_assign($1, double(*$1) * $3);
			   }
	| VAR SLASHE exp { $$ = maybe_assign($1, double(*$1) / $3); }
	| VAR PERCENTE exp { $$ = maybe_assign($1,
					       /*(double)*/
					       ((int) *$1 %
						(int) $3)); }
	| FNCT '(' exp ')'{ $$ = (*($1->fnctptr))($3); }
	| exp '+' exp	{ $$ = $1 + $3; }
	| exp '-' exp   { $$ = $1 - $3; }
	| exp '*' exp   { $$ = $1 * $3; }
	| exp '/' exp   { $$ = $1 / $3; }
	| exp '%' exp   { $$ = (int) $1 % (int) $3; }
	| PLUSPLUS VAR	{ $$ = *$2;
			  maybe_assign($2, double(*$2) + 1.0); }
	| MINUSMINUS VAR { $$ = *$2;
			   maybe_assign($2, double(*$2) - 1.0); }
	| VAR PLUSPLUS	{ $$ = double(*$1) + 1.0;
			       maybe_assign($1, double(*$1) + 1.0); }
	| VAR MINUSMINUS { $$ = double(*$1) - 1.0;
			       maybe_assign($1, double(*$1) - 1.0); }
	| '-' exp  %prec UMINUS { $$ = -$2; }
	| '+' exp  %prec UPLUS  { $$ = +$2; }
	| '!' exp       { $$ = ! (int) $2; }
	| '~' exp       { $$ = ~ (int) $2; }
	| exp '^' exp   { $$ = pow($1, $3); }
	| '(' exp ')'   { $$ = $2; }
	| exp LL exp    { $$ = (int) $1 << (int) $3; }
	| exp GG exp    { $$ = (int) $1 >> (int) $3; }
	| exp EQ exp    { $$ = $1 == $3; }
	| exp NE exp    { $$ = $1 != $3; }
	| exp LE exp    { $$ = $1 <= $3; }
	| exp GE exp    { $$ = $1 >= $3; }
	| exp '<' exp   { $$ = $1 < $3; }
	| exp '>' exp   { $$ = $1 > $3; }
	| exp '&' exp   { $$ = (int) $1 & (int) $3; }
	| exp '|' exp   { $$ = (int) $1 | (int) $3; }
	| exp AND exp   { $$ = (int) $1 && (int) $3; }
	| exp OR exp    { $$ = (int) $1 || (int) $3; }
	| exp '?' exp ':' exp 
			{ $$ = ((int) $1 ? $3 : $5); }
	| SAVE_STATE '(' VAR ')'
			{ $$ = save_state($3->name); }
	| SAVE_STATE '(' VAR ',' exp ')'
			{ $$ = save_state_numbered($3->name,
						   (unsigned int) $5); }
;
%%
