/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

/* $Id: error.cc,v 1.13 1994/02/14 22:24:02 bengt Exp $ */

/*
 * This file contains an error reporting function with variable number of
 * arguments.
 */

#include "externals.h"
#include <stdarg.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>

#define REENTRANT

#ifndef REENTRANT
static char mess[10*MAXSTRINGSIZE];
static char logmessage[10*MAXSTRINGSIZE];
#endif

/* Forward declaration */
static char *sev_str(severity sev);

// To prevent from loops when error occurs in emergency_dump
static bool fatal_in_progress = false;

/*
 * General purpose error/warning/message reporting function
 * First argument is the severity of the "error", the rest of the arguments
 * are like printf.
 */
extern void 
gnans_error(severity sev, const char *fmt,...)
{
#ifdef REENTRANT
  char mess[10*MAXSTRINGSIZE];
  char logmessage[10*MAXSTRINGSIZE];
#endif
  int pid;
  va_list args;
  va_start(args, fmt);
  vsprintf(mess, fmt, args);
  if (sev != nonerror_message && sev != debug) {
    abort_script = running_script;
    running_integration = false;
  }

  if (do_logging) {
    sprintf(logmessage,
	    "gnans_error, severity = %s, ``%s''", sev_str(sev), mess);
    log_event(logmessage);
  }

  switch (sev) {
  case non_fatal:
  case recoverable:
    if (X) 
      popup_error_message(toplevel, NULL, mess, "Error", XtGrabNone);
    else
      fprintf(stderr, "%s: %s\n",
	      (current_system.system_name.empty() ? progname : current_system.system_name).c_str(),
	      mess);
    break;
  case terminate_simulation:
    if (X) 
      popup_error_message(toplevel, NULL, mess, "Simulation Terminated",
			  XtGrabNone);
    else
      fprintf(stderr, "%s: %s\n",
	      (current_system.system_name.empty() ? progname : current_system.system_name).c_str(),
	      mess);
    break;
  case message:
  case nonerror_message:
    if (X) 
      popup_error_message(toplevel, NULL, mess, "Gnans Message", XtGrabNone);
    else
      fprintf(stderr, "%s[%s]: %s\n", progname.c_str(),
	      (current_system.system_name.empty() 
	       ? "" : current_system.system_name).c_str(),
	      mess);
    break;
  case warning:
    if (X) 
      popup_error_message(toplevel, NULL, mess, "Gnans Warning", XtGrabNone);
    else
      fprintf(stderr, "%s: %s\n",
	      (current_system.system_name.empty() 
	       ? "" : current_system.system_name).c_str(),
	      mess);
    break;
  case just_say_no:
  case lock_exists:
    fprintf(stderr, "%s: %s\n", progname.c_str(), mess);
    die("", sev);
    break;
  case fatal:
    if (! fatal_in_progress) {
      fatal_in_progress = true;
      fprintf(stderr, "%s[%s]: %s\n", progname.c_str(),
	      (current_system.system_name.empty() 
	       ? "" : current_system.system_name).c_str(),
	      mess);
      emergency_dump();
    }
    die("This error is fatal", fatal);
    break;
  case signal_handler_message:
    fprintf(stderr, "%s[%s]: %s\n", progname.c_str(),
	    (current_system.system_name.empty() 
	     ? "" : current_system.system_name).c_str(),
	    mess);
    break;
  case impossible:
  case run_fatal:
    if (! fatal_in_progress) {
      fatal_in_progress = true;
      fprintf(stderr, "%s: %s\n",
	      (current_system.system_name.empty() 
	       ? "" : current_system.system_name).c_str(),
	      mess);
      perror("Perror: ");
      emergency_dump();
    }
    abort();
    break;
  case coredump:
    /*
     * First fork, then abort. Thus I get a core dump while the program
     * survives.
     */
    if (X) 
      popup_error_message(toplevel, NULL, mess, "Core Dump", XtGrabNone);
    else
      fprintf(stderr, "%s: %s\n",
	      (current_system.system_name.empty() 
	       ? "" : current_system.system_name).c_str(),
	      mess);

    pid = fork();
    if (pid > 0)		/* Parent */
      ;
    else if (pid == 0)		/* Child */
      abort();
    else {			/* Fork failed */
      fprintf(stderr, "gnans: Fork failed, quitting with core dump\n");
      close_log();
      abort();
    }
    break;
  case parse:
    fprintf(stderr, "%s[%s]: %s\n", progname.c_str(),
	    current_system.system_name.c_str(), mess);
    break;
  case debug:
    fprintf(stdout, "%s[%s]: %s\n", progname.c_str(),
	    current_system.system_name.c_str(), mess);
    break;
  default:
    fprintf(stderr, "%s\n", mess);
    fprintf(stderr, "%s: Internal error, dumping core\n",
	    progname.c_str());
    close_log();
    abort();
    break;
  }
}

/*
 * Utility function converting the severity code to a text string.
 */ 
static char *
sev_str(severity sev)
{
  switch(sev) {
  case impossible:
    return "impossible";
  case fatal:
    return "fatal";
  case run_fatal:
    return "run_fatal";
  case non_fatal:
    return "non_fatal";
  case just_say_no:
    return "just_say_no";
  case lock_exists:
    return "lock_exists";
  case parse:
    return "parse";
  case semantic:
    return "semantic";
  case recoverable:
    return "recoverable";
  case debug:
    return "debug";
  case warning:
    return "warning";
  case message:
    return "message";
  case nonerror_message:
    return "nonerror_message";
  case coredump:
    return "coredump";
  case signal_handler_message:
    return "signal_handler_message";
  case terminate_simulation:
    return "terminate_simulation";
  }
  return "impossible";
}
