/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

/* $Id: fileselector.cc,v 1.14 1994/11/01 12:52:38 bengt Exp bengt $ */

/*
 * As the name suggests, this file should contain a file selector. It
 * doesn't. Yet. But the programming interface will be the same.
 */


#include "externals.h"
#include <errno.h>
#include <stdlib.h>
#include <sys/stat.h>

#define String XtString
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>

#ifdef USEXAW3DINCLUDES

#include <X11/Xaw3d/Box.h>
#include <X11/Xaw3d/Form.h>
#include <X11/Xaw3d/Cardinals.h>
#include <X11/Xaw3d/Command.h>
#include <X11/Xaw3d/Dialog.h>
#include <X11/Xaw3d/List.h>

#else

#include <X11/Xaw/Box.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/List.h>

#endif

extern "C" char *tilde_expand(char *); /* From Gnu readline */

#define STR_EQUAL(s1, s2) strcmp((s1), (s2)) == 0
#define widgetlog if (do_logging && trace_widgets) va_log_event

/*
 * I need this global garbage since actions do not have client_datas.
 * It makes the fileselector non-reentrant, however this is not a problem
 * if we grab all events (which also makes sense for the user).
 */

/* Structure needed to pass data */
typedef struct  _fileselector_call {
  void (*run_fcn)(char *str);
  void (*message_fcn)(char *str);
  char *string;
  bool empty_ok;
  bool must_exist;
} fileselector_call;

static fileselector_call *global_fileselector_data;

static void
start_work(XtPointer client_data, XtIntervalId */*id*/)
{
  fileselector_call *call = (fileselector_call *) client_data;
  (*call->run_fcn)(call->string);
}

static void 
fileselector_open(Widget mother_widget, XtPointer client_data,
		  XtPointer /*call_data*/)
{
  Widget w;
  char str[MAXSTRINGSIZE];
  int result;
  typedef struct stat struct_stat;
  struct_stat *buf = new struct_stat;
  fileselector_call *call = (fileselector_call *) client_data;
  
  strcpy(str, (char *) XawDialogGetValueString(XtParent(mother_widget)));
  char *expanded_name = tilde_expand(str);
  strcpy(str, expanded_name);
  free(expanded_name);
  widgetlog("fileselector_open, filename = ``%s''", str);

  /* Empty? */
  if (STR_EQUAL(str, "")) {
    if (! call->empty_ok) {
      gnans_error(warning, "Empty filename not accepted");
      return;
    }

  } else {

    // If it ends with a ``|'', knock it off before testing
    char teststr[MAXSTRINGSIZE];
    strcpy(teststr, str);
    if (teststr[strlen(str)-1] == '|')
      teststr[strlen(str)-1] = '\0';

    /* Does it exit? */
    result = stat(teststr, buf);
    if (result) {
      switch (errno) {
      case ENOENT:
	if (call->must_exist) {
	  gnans_error(warning, "Non-existent file ``%s''", str);
	  return;
	}
	break;
      case ENOTDIR:
	gnans_error(warning,
		    "A component of the path prefix of ``%s'' "
		    "is not a directory",
		    str);
	return;
	break;
      case EACCES:
	gnans_error(warning,
		    "Search permission is denied for a component of the "
		    "path prefix of ``%s''",
		    str);
	return;
	break;
      case ELOOP:
      case ENAMETOOLONG:
	gnans_error(warning, "Illegal file name ``%s''", str);
	return;
	break;
      }
    }
  }

  /* This is not the right place to check if the file is readable. */
  strcpy(call->string, str);

  w = XtParent(XtParent(mother_widget));
  XtDestroyWidget(w);
  /*  XSync(XtDisplay(toplevel), False);*/
  (*call->message_fcn)(call->string);
  XtAppAddTimeOut(XtWidgetToApplicationContext(mother_widget),
                  500, &start_work, (XtPointer) call);

  delete buf;
}

extern void
fileselector_open_action(Widget w, XEvent */*event*/,
			 String */*params*/, Cardinal */*num_params*/)
{
  fileselector_open(w, (XtPointer) global_fileselector_data, NULL);
}

#define NOT_SELFILE
#ifdef SELFILE
extern "C" FILE *
XsraSelFile(Widget toplevel,
	    char            *prompt,
	    char            *ok,
	    char            *cancel,
	    char            *failed,
	    char            *init_path,
	    char            *mode,
	    boolean *show_it(char *fileName, char **retFileName, struct stat *statbuf),
	    char            **name_return);
extern "C" FILE *
SelFile(Widget toplevel,
	char *application_title, /* Application title of toplevel widget */
	char *application_name,	/* Application name of toplevel widget */
	char *init_path,
	int *f(char *fileName, char **retFileName, struct stat *statbuf),
	int open_it,		/* Should it be opened? */
	char *mode,		/* Mode to open */
	char **name_return,	/* returns the name */
	boolean *success);		/* Did the selection succeed? */

static bool
gnans_extension(char *fileName, char **retFileName, struct stat *statbuf)
{
  int extIdx;
  char *fileselector_ending = ".gnans";

  *retFileName = fileName;
  extIdx = strlen(fileName) - strlen(fileselector_ending);
  return(S_ISDIR (statbuf->st_mode) || 
	 ((extIdx > 0) && STR_EQUAL(fileName + extIdx, fileselector_ending)));
}
#endif

extern void
fileselector(char *title, char *name_of_label_widget, char */*ending*/,
	     char *filename_ptr, void (*run_fcn)(char *str),
	     void (*message_fcn)(char *str),
	     bool must_exist, bool empty_ok)
{
#ifdef SELFILE
  char *filename;
  boolean success;
  SelFile(toplevel, title, name_of_label_widget, NULL,
		 STR_EQUAL(ending, ".gnans") ? &gnans_extension : NULL,
		 false, NULL, &filename , &success);
#if 0
		     NULL,
		     NULL,
		     NULL,
		     NULL,
		     "r",
		     STR_EQUAL(ending, ".gnans") ? &gnans_extension : NULL,
		     &filename;
#endif
  if (success)
    gnans_error(debug, "You selected %s", filename);
  else
    gnans_error(debug, "No success");
  
#else /* ! SELFILE */
  Widget filenamerequester, dialog, button;
  char ttl[MAXSTRINGSIZE];
  fileselector_call *call_ptr = new fileselector_call;

  global_fileselector_data = call_ptr;
  call_ptr->run_fcn = run_fcn;
  call_ptr->message_fcn = message_fcn;
  call_ptr->string = filename_ptr;
  call_ptr->must_exist = must_exist;
  call_ptr->empty_ok = empty_ok;
  
  make_popup_title(ttl, title),
  filenamerequester = XtVaCreatePopupShell("filenamerequester",
					   transientShellWidgetClass,
					   toplevel,
					   XtNtitle, ttl,
					   NULL);
    
  dialog = XtVaCreateManagedWidget(name_of_label_widget, dialogWidgetClass,
				   filenamerequester,
				   XtNvalue, filename_ptr, NULL);
    
  button =  XtVaCreateManagedWidget("open", commandWidgetClass, dialog, NULL);
  XtAddCallback(button, XtNcallback, fileselector_open, (XtPointer) call_ptr);
  button =  XtVaCreateManagedWidget("cancel", commandWidgetClass,
				    dialog, NULL);
  XtAddCallback(button, XtNcallback, destroy_inquiry_popup,
		(XtPointer) &filenamerequester);
  popup_position(filenamerequester, toplevel, XtGrabNone, false);
#endif
}
