/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

// $Id: integrate.cc,v 1.29 1994/08/19 18:38:31 bengt Exp $

// This file contains routines for setting up systems, static or dynamic,
// and routines for starting integrations.

#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <errno.h>
#include <unistd.h>
//#include <iostream>

#include "externals.h"
#include "paths.h"
#include "utillib.h"
#include "CIntegrator.h"

typedef char *String;		// ???????????????????????? FIXME
#include "Plotter.h"
#include "PlotterControl.h"
#include "Store.h"

#define log if (do_logging) va_log_event

#ifdef DYNAMIC_LOADING
#include <dlfcn.h>
#endif				// DYNAMIC_LOADING

// True if I have not Start'ed the system (thus Continue meaningless)
static bool virgin_system = true;

// Setting this true forces initialization of trajectories
// in the next integration
static bool need_initialization = false;

static void delete_stores();

// Allocates doubles, with error control.
static double *
dmalloc(unsigned int n)
{
  double *ptr = (double *) malloc(n * sizeof(double));
  if (ptr == NULL) {
    gnans_error(fatal, "Cannot allocate %d more bytes.", n * sizeof(double));
    return NULL;
  } else {
    return ptr;
  }
}

// Frees space allocated with dmalloc.
static void 
dfree(double *p)
{
  if (malloc_debug)
    gnans_error(debug, "dfree called with argument %x", (long int) p);
  free(p);
}

// Like realloc, but for doubles and with error control. Equal to dmalloc
// if called with NULL argument.
static double *
drealloc(double *old, unsigned n)
{
  double *ptr;
  unsigned int i;
  if (malloc_debug)
    gnans_error(debug, "Realloc called with n = %d; old = %x",
		n, (long int) old);
  if (! old)
    return dmalloc(n);
  
  ptr = (double *) realloc(old, n * sizeof(double));

  if (ptr && malloc_debug)
    gnans_error(debug, "Realloc succeeded; new = %x", (long int) ptr);
  
  if (ptr == NULL) {
    if (errno == EINVAL)
      gnans_error(warning,
		  "Error EINVAL by realloc, doing it the primitive way");
    else if (errno == ENOMEM)
      gnans_error(warning,
		  "Error ENOMEM by realloc, doing it the primitive way");
    else
      gnans_error(warning,
		  "Cannot reallocate memory, doing it the primitive way");
    ptr = dmalloc(n);
    for (i = 0; i < n - MALLOCUNIT; i++)
      ptr[i] = old[i];
    dfree(old);
  }
  return ptr;
}
/***************************************************************/
static bool 
initialize_run()
{
  delete_stores();
  // Compute number_of_stores and allocate space
  number_of_stores = 0;
  unsigned int i;
  for (i = 0; i < number_of_internals; i++) // t and so on...
    if (internals[i]->store_me())
      number_of_stores++;
  for (i = 0; i < current_system.number_of_states; i++)
    if (current_system.states[i]->store_me())
      number_of_stores++;
  for (i = 0; i < current_system.number_of_functions; i++) 
    if (current_system.functions[i]->store_me())
      number_of_stores++;
  for (i = 0; i < current_system.number_of_parameters; i++) 
    if (current_system.parameters[i]->store_me())
      number_of_stores++;
  for (i = 0; i < current_system.number_of_randoms; i++) 
    if (current_system.randoms[i]->store_me())
      number_of_stores++;
  stores = new Store[number_of_stores];
  
  unsigned int store_index = 0;
  for (i = 0; i < number_of_internals; i++) // t and so on...
    if (internals[i]->store_me()) {
      stores[store_index].name = internals[i]->name();
      stores[store_index++].ptr = *internals[i];
    }
  for (i = 0; i < current_system.number_of_states; i++)
    if (current_system.states[i]->store_me()) {
      stores[store_index].name = current_system.states[i]->name();
      stores[store_index++].ptr =
	(double *) current_system.state_vector + i;
    }
  for (i = 0; i < current_system.number_of_functions; i++) 
    if (current_system.functions[i]->store_me()) {
      stores[store_index].name =
	current_system.functions[i]->name();
      stores[store_index++].ptr =
	*current_system.functions[i];
    }
  for (i = 0; i < current_system.number_of_parameters; i++) 
    if (current_system.parameters[i]->store_me()) {
      stores[store_index].name =
	current_system.parameters[i]->name();
      stores[store_index++].ptr =
	*current_system.parameters[i];
    }
  for (i = 0; i < current_system.number_of_randoms; i++) 
    if (current_system.randoms[i]->store_me()) {
      stores[store_index].name =
	current_system.randoms[i]->name();
      stores[store_index++].ptr =
	*current_system.randoms[i];
    }

  // Check consistency in what is malloc-ed for the trajectories
  if (trajectories_current_size > 0)
    for (i = 0; i < number_of_stores; i++)
      if (! stores[i].trajectory)
	stores[i].trajectory = dmalloc(trajectories_current_size);
  
  if (! append_trajectories || need_initialization) {
    sample_no = 0;
    Store::reset_discontinuities();
  }
  if (trajectories_current_size == 0) {
    for (i = 0; i < number_of_stores; i++)
      stores[i].trajectory = dmalloc(MALLOCUNIT);
    trajectories_current_size = MALLOCUNIT;
  }
  need_initialization = false;
  return true;
}

extern void
resize_trajectories()
{
  trajectories_current_size += MALLOCUNIT;
  for (unsigned i = 0; i < number_of_stores; i++)
    stores[i].trajectory = drealloc(stores[i].trajectory,
				    trajectories_current_size);
}

static void
delete_stores()
{
  if (stores) {
    for (unsigned long int j = 0; j < number_of_stores; j++) {
      dfree(stores[j].trajectory);
      stores[j].trajectory = NULL;
    }
    delete [] stores;
  }
  trajectories_current_size = 0;
  stores = NULL;
}

// Called at the end of a run. Writes all trajectories to disk
extern void
write_trajectories()
{
  // Ignore empty trajectories
  if (sample_no == 0)
    return;

  if (std::string(store_format_table[store_format].name) == "Null")
    return;

  enable_busy_window();
    
  // Open
  std::string store_filename = GNANSSTOREFILENAME;
  store_filename += "." + std::string(store_format_table[store_format].extension);

  if (! (stores[0].file = fopen(store_filename.c_str(),"w"))) {
    gnans_error(non_fatal, "Cannot open file %s", store_filename.c_str());
    return;
  }
    
  // Write with the format supplied function
 if (! (*store_format_table[store_format].write)())
   gnans_error(non_fatal, "Write failed (disk or quota full?)");
  
  // Then close
  if (fclose(stores[0].file))
    gnans_error(non_fatal, "Cannot close file");
  
  disable_busy_window();
}

// This is called when flipping append_trajectories
extern void
append_trajectories_hook()
{
  write_trajectories();
  need_initialization = true;
}

/****************************************************************/
// This function starts the integration. If using workprocs,
// it basically initializes the workprocs.
// Returns true if the integration was completed or interrupted by the user 
// or if error occured (i.e. we are "finished").
// (Actually, this is presently never used...)
// The argument says if it is a continuation run or not. If it is, then
// the initial values are not reset. If systemtime != stoptime, systemtime is
// reset, otherwise not.
extern bool
integrate(Continuation continuation)
{
  // Use this to indicate if the user has changed algo since pressing
  // the "Stop" button.
  static Integrator *last_algorithm = NULL;
  if (virgin_system && continuation != init_all) {
    gnans_error(non_fatal,
		"``Continue'' without first ``Start''-ing is not possible");
    return true;
  }

  if (continuation != init_all && floating_point_error_occured) {
    gnans_error(non_fatal, "Continuation not possible since last run "
		"ended with a floating point error");
    return true;
  }
  
  if (timestep < 0.0) {
    gnans_error(non_fatal, "Timestep must be nonnegative");
    return true;
  }

  if (continuation != init_all && storeset_dirty) {
    gnans_error(non_fatal,
		"Continuation not possible, since the store "
		"status of some variables has changed since last run.");
    return true;
  }
  
  if (continuation != init_all && current_algorithm != last_algorithm) {
    gnans_error(non_fatal,
		"Continuation not possible, since the integration "
		"algorithm has changed since last run.");
    return true;
  }
  
  // Check that the selected format makes sense
  if (! (*store_format_table[store_format].check)(true)) {
    return true;
  }

  // Reset systemtime UNLESS this is a continuation run AND 
  // systemtime < stoptime (systemtime > stoptime if running backwards).
  if (continuation != init_all &&
      (starttime < stoptime ?
       systemtime < stoptime - timestep :
       systemtime > stoptime + timestep)) {
    // This is a true continuation, do nothing
    if (trace_calls)
      gnans_error(debug, "Continuation run started");
  } else {
    // Reset system time
    systemtime = double(starttime);
    if (trace_calls)
      gnans_error(debug, "Initialization run started");
  }
  
  // perform timeinvariant computations
  // If it fails, the system supplied function should report the error
  if ((*current_system.init_ptr)(current_system.state_vector,
				 continuation,
				 current_algorithm->stochastic())) {
    Plotter::disable_plotters();
    if (storeset_dirty)
      PlotterControl::destroy_plotter_controls();
    storeset_dirty = false;
    if (sample_no == 0 || ! append_trajectories || need_initialization)
      if (! initialize_run()) {
	// gnans_error(non_fatal, "Error occured by initialization");
	return true;
      }
    
    Store::add_discontinuity();

    virgin_system = false;
    running_integration = true;
    floating_point_error_occured = false;
    last_algorithm = current_algorithm;
    log("Integrate from %g to %g%s", double(systemtime), double(stoptime),
	continuation == init_none ? " (Continuation)"
	: (continuation == init_states ? " (Continuation_init)" : ""));
    if (X && ! running_script && workprocs) {
      enable_buttons_busy();
      register_work_proc(systemtime != starttime);
      return true;
    } else
      return continue_integration(systemtime == starttime);
  }
  return true;
}

// This function calls the integrator.
// Returns true if the integration was completed or interrupted by the user.
extern bool
continue_integration(bool initialize)
{
  bool finished;
  
  finished = current_algorithm->loop(initialize);
  if (finished) {
    running_integration = false;
    integrate_finish();
    if (! workprocs && X && ! running_script)
      indicate_progress();
    return true;
  } else
    // More to do
    return false;
}

// Finish up the integration
extern void
integrate_finish()
{
  (*current_system.finish_ptr)();
  log_event("");		// Flushes logfile
  if (! append_trajectories) {
    write_trajectories();
  }
}

//***************************************************************/
// This function sets up a static system, i.e. using a system present in
// the executable at linktime (the ACSL way)
extern void
setup_system_static()
{
  current_system.init_ptr = &gnans_init;
  current_system.finish_ptr = &gnans_finish;
  current_system.setup_ptr = &gnans_setup;
  current_system.f_ptr = &gnans_f;
  current_system.discrete_f_ptr = &gnans_discrete_f;
  current_system.f_store_ptr = &gnans_f_store;
  current_system.f_accepted_ptr = &gnans_f_accepted;
  current_system.destroy_ptr = &gnans_destroy;

  bool success = (*current_system.setup_ptr)(current_system.state_vector,
						systemtime);
  if (!success)
    // This is a serious problem, continuation is meaningless
    gnans_error(impossible, "Failure in setup");

  current_system.state_vector.resize(current_system.number_of_states);

  if (X) {
    set_toplevel_title();
    disable_some_buttons();
  }
  virgin_system = true;
  
  switch(current_system.system_type) {
  case continuous:
  case hybrid:
    if (current_system.number_of_wieners)
      select_algorithm(default_stochastic_algorithm.c_str(), true);
    else
      select_algorithm(default_algorithm.c_str(), true);
    break;
  case discrete:
    select_algorithm((Integrator*) &discrete_time_integrate);
    timestep = 0;
    break;
#ifdef DAS
  case das:
    gnans_error(impossible, "Not implemented");
    //select_algorithm(&dasl_integrate);
    break;
#endif
#ifdef RICCATI
  case riccati:
    select_algorithm((Integrator*) &riccati_integrate);
    break;
#endif
  default:
    gnans_error(impossible, "Programming error");
    break;
  }
}

/****************************************************************/
#ifdef DYNAMIC_LOADING
#define DLSYMBOL(ptr, handle, symbol, type) ptr = ( type ) dlsym(handle, symbol); \
  if (!ptr) {\
               gnans_error(non_fatal, "Could not find symbol %s", symbol);\
                 return;\
                 }

static void *handle = NULL;

typedef void (*void_void_t)();
typedef int (*setup_t)(double **state_vector_ptr, double *system_time);
typedef int (*init_t)(double state_vector[],
		      Continuation continuation,
		      bool stochastic);
typedef void (*f_t)(double *time_ptr,
		    double state_vector[1],
		    double d_states[1]);

static void
getpath(std::string &var, const char *environment_name, const char *def)
{
  var = getenv(environment_name) ? getenv(environment_name) : def;
}

// This function compiles and loads dynamically the system with its file
// name as argument
extern void
setup_system_dynamic(std::string system_filename)
{
  std::string so_filename;
  std::string compile_command;
  int return_status;

  if (system_filename == "") {
    gnans_error(message, "System unchanged");
    return;
  }

  if (system_filename[system_filename.length()-1] == '.')
    system_filename += "gnans";
  if (system_filename.find(".gnans") == std::string::npos)
    system_filename += ".gnans";

  typedef struct stat struct_stat;
  struct_stat *buf = new struct_stat;
  int result = stat(system_filename.c_str(), buf);
  delete buf;
  if (result) {
    switch (errno) {
    case ENOENT:
      gnans_error(message, "Non-existent file ``%s''", system_filename.c_str());
      break;
    case ENOTDIR:
      gnans_error(message,
		  "A component of the path prefix of ``%s'' "
		  "is not a directory",
		  system_filename.c_str());
      break;
    case EACCES:
      gnans_error(message,
		  "Search permission is denied for a component of the "
		  "path prefix of ``%s''",
		  system_filename.c_str());
      break;
    case ELOOP:
    case ENAMETOOLONG:
      gnans_error(message, "Illegal file name ``%s''", system_filename.c_str());
      break;
    }
    return;
  }

  delete_stores();
  //so_filename = system_filename;
  so_filename = system_filename.substr(0, system_filename.length() - 5) +  SHARED_OBJECT_ENDING;
  // If it is a relative pathname, prepend ``./'' in order not to impose
  // any assumptions on the run time path for the loader
  if (so_filename[0] != '/' && so_filename[0] != '.')
  so_filename = "./" + so_filename;

  // This is a kludge, should be done differently later.
  std::string cflags;
  if (getenv("GNANSCOMPILECFLAGS")) {
    cflags = " CFLAGS='";
    cflags += "' ";
    cflags += getenv("GNANSCOMPILECFLAGS");
  }
  std::string gnanslator_cmd, gnans_includes, gnans_makefile;
  getpath(gnanslator_cmd, "GNANSLATOR", GNANSLATOR);
  getpath(gnans_includes, "GNANSINCLUDES", GNANSINCLUDES);
  getpath(gnans_makefile, "GNANSMAKEFILE", GNANSMAKEFILE);
  compile_command =
    "make GNANSLATOR=" + gnanslator_cmd + cflags
    + " GNANSINCLUDES=-I" + gnans_includes 
    + " -f " + gnans_makefile + " " + so_filename;

  log("Trying to compile with command ``%s''", compile_command.c_str());

  std::string error_message;
  int from_compilation[2];
  pipe(from_compilation);
  pid_t pid = fork();
  if (pid > 0) {		// Parent
    int from_compilation_fd = dup(from_compilation[0]);
    FILE *f = fdopen(from_compilation_fd, "r");
    close(from_compilation[0]);
    close(from_compilation[1]);
    int c;
    const int max_error_message_length = 3000;
    while (c = fgetc(f), !feof(f))
      if (error_message.length() < max_error_message_length)
	error_message += (char) c;
    if (error_message.length() == max_error_message_length)
      error_message += " ...";
#ifdef USE_WAITPID
    return_status = 0;
    errno = 0;
    pid_t status = waitpid(pid, &return_status, 0);
    if (trace_calls)
      gnans_error(debug, "status = %d, errno = %d", status, errno);
    if (status == -1)
      gnans_error(fatal, "Problems in waitpid, errno = %d", errno);
    if (errno == EINTR) {
      gnans_error(non_fatal, "Compilation aborted due to signal.\n"
		  "Loaded system unchanged.");
      return;
    }
#else
    while (wait(&return_status) != pid)
      ;
#endif
  } else if (pid == 0) {	// Child
    dup2(from_compilation[1], 1); // Redirect stdout
    dup2(from_compilation[1], 2); // Redirect stderr
    close(from_compilation[0]);
    close(from_compilation[1]);
    setbuf(stdout, NULL);
    setbuf(stderr, NULL);
    execl("/bin/sh", "/bin/sh", "-c", compile_command.c_str(), NULL);
  } else {
    gnans_error(non_fatal, "Problem by forking");
    return;
  }

  if (error_message.length() > 1000)
    error_message = error_message.substr(0,1000) + std::string("...");
  if (return_status) {
    gnans_error(non_fatal,
		"Compilation failed. Loaded system unchanged. "
		"Error messages:\n"
		"=======================================================\n"
		"%s", error_message.c_str());
    return;
  }
  if (handle) {
    (*current_system.destroy_ptr)();
#ifdef BROKEN_DLCLOSE
    gnans_error(warning,
		"This version of Gnans was compiled for not using\n"
		"dlclose, which is supposed to be broken on this system.\n"
		"If the new system has the same file name as a previously\n"
		"loaded system, CHANGES WILL NOT TAKE EFFECT!");
#else
    if (dlclose(handle)) {
      gnans_error(impossible, "dlclose failed");
    }
#endif
  }
  handle = dlopen(so_filename.c_str(), RTLD_NOW);
  if (! handle) {
    gnans_error(non_fatal, "dlopen failed: %s", dlerror());
    return;
  }
  DLSYMBOL(current_system.setup_ptr, handle, "gnans_setup", setup_t);
  DLSYMBOL(current_system.init_ptr, handle, "gnans_init", init_t);
  DLSYMBOL(current_system.f_ptr, handle, "gnans_f", f_t);
  DLSYMBOL(current_system.finish_ptr, handle, "gnans_finish", void_void_t);
  DLSYMBOL(current_system.discrete_f_ptr, handle, "gnans_discrete_f",
	   void_void_t);
  DLSYMBOL(current_system.f_accepted_ptr, handle, "gnans_f_accepted",
	   void_void_t);
  DLSYMBOL(current_system.f_store_ptr, handle, "gnans_f_store",
	   void_void_t);
  DLSYMBOL(current_system.destroy_ptr, handle, "gnans_destroy",
	   void_void_t);

  log("Set up a system from file ``%s''", system_filename.c_str());

  sample_no = 0;
  number_of_stores = 0;
  
  bool success = (*current_system.setup_ptr)(current_system.state_vector,
					     systemtime);
  if (! success)
    // The setup function should report the error, not we.
    return;

  current_system.state_vector.setsize(current_system.number_of_states);

  if (X) {
    zap_popups();
    set_toplevel_title();
    disable_some_buttons();
    Plotter::disable_plotters();
  }
  
  virgin_system = true;

  switch(current_system.system_type) {
  case continuous:
  case hybrid:
    if (current_system.number_of_wieners) {
      if (! current_algorithm->stochastic())
	select_algorithm(default_stochastic_algorithm.c_str());
    } else
      if (current_algorithm->stochastic()
	  || current_algorithm == (Integrator*) &discrete_time_integrate
#ifdef RICCATI
	  || current_algorithm == (Integrator*) &riccati_integrate
#endif
	  )
	select_algorithm(default_algorithm.c_str());
    break;
  case discrete:
    select_algorithm((Integrator*) &discrete_time_integrate);
    timestep = 0;
    break;
#ifdef DAS
  case das:
    select_algorithm(&dasl_integrate);
    break;
#endif
#ifdef RICCATI
  case riccati:
    select_algorithm((Integrator*) &riccati_integrate);
    break;
#endif
  default:
    gnans_error(impossible, "Programming error");
    break;
  }
}
#endif				// DYNAMIC_LOADING
