/* -*-c-*- */
/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

/* $Id: lexical.l,v 1.16 1994/11/01 12:52:38 bengt Exp bengt $ */

/*
 * This file contains the lexical analyser for the script language.
 * It REQUIRES flex.
 */

%{
/* We take advantage of the states in flex */
#ifndef FLEX_SCANNER
#error Sorry, flex, not lex, is required
#endif

/* Must come before ``include "externals.h"'', sorry */
extern "C" char * readline(const char *);
extern "C" void add_history(const char *);

#include "externals.h"
#include "cmdparse.h"

#ifndef M_PI			/* if __STRICT_ANSI__ is in effect... */
#define M_PI            3.14159265358979323846
#endif

#ifdef READLINE
#undef YY_INPUT
#define YY_INPUT(buf,result,max_size) my_yy_input(buf, result, max_size)

typedef int Function();
/* A static variable for holding the line. */
static char *line_read = 0;

/* Here is a bug: If running in the background, the terminal modes are
   not reset, so the user input is not echoed. */

/* Read a string, and return a pointer to it.  Returns NULL on EOF. */
char *
do_gets()
{
  /* If the buffer has already been allocated, return the memory
     to the free pool. */
  if (line_read) {
    free(line_read);
    line_read = 0;
  }

#if 1
  /* Get a line from the user. */
  line_read = readline(cmd_prompt.c_str());
  
  /* If the line has any text in it, save it on the history. */
  if (line_read && *line_read)
    add_history(line_read);
  
#else
  line_read = (char *) malloc(1000);
  gets(line_read);
#endif

  return(line_read);
}

static void
my_yy_input(char *buf, int &result, unsigned int max_size)
{
  if (yyin == stdin && isatty(fileno(stdin)) && isatty(fileno(stdout))) {
    char *ans = do_gets();
    if (!ans) {
      result = YY_NULL;
      return;
    }
    if (strlen(ans) > max_size - 1)
      gnans_error(fatal, "too long");
    strcpy((char *) buf, ans);
    strcat((char *) buf, "\n");
    result = strlen((char *) buf);
  } else
    if (! fgets((char *) buf, max_size, yyin)) {
      result = YY_NULL;
      return;
    }
  result = strlen((char *) buf);
}
#endif /* Readline */

// This array contains all the name of the commands, and is used for the 
// readline completer.
char * script_commands[] = {
  "start",
  "run",
  "quit",
  "exit",
#ifdef HELP
  "help",
#endif
  "display",
  "store",
  "unstore",
  "switch",
  "system",
  "continue",
  "continue_init",
  "set",
  "save_state(",
  "restore_state(",
  "plot",
  NULL,
};

%}

%option noyywrap

DIGITS	([0-9])+
FLOAT	({DIGITS}?("."{DIGITS})?)|({DIGITS}("."{DIGITS})?[eE][+-]?{DIGITS})|(Inf(inity)?)
HEX     0x([0-9a-fA-F])+
OCT     0([0-7])+
ID	[a-zA-Z]([a-zA-Z0-9_])*"@"?
STRING 	\"[^\"]*\"
SHELLCMD "!"[^\n]+
COMMENT "#"[^\n]*

%%
start 		{ return START; }
run   		{ return START; }
quit  		{ return QUIT; }
exit  		{ return QUIT; }
help  		{ return HELP; }
display 	{ return DISPLAY; }
set	 	{ return SET; }
store		{ return STORE; }
unstore		{ return UNSTORE; }
switch		{ return SWITCH; }
system		{ return SYSTEM; }
continue	{ return CONTINUE; }
continue_init	{ return CONTINUE_INIT; }
algorithm	{ return ALGORITHM; }
format		{ return FORMAT; }
plot 		{ return PLOT; }
axis 		{ return AXIS; }
true 		{ return GNANS_TRUE; }
false 		{ return GNANS_FALSE; }
lineplot	{ return LINEPLOT; }
linlin 		{ return LINLIN; }
linlog 		{ return LINLOG; }
loglin 		{ return LOGLIN; }
loglog 		{ return LOGLOG; }

save_state	{ return SAVE_STATE; }
restore_state   { return RESTORE_STATE; }

on		{ return ON; }
off		{ return OFF; }

functions	{ return FUNCTIONS; }
internals	{ return INTERNALS; }
resources	{ return RESOURCES; }
parameters	{ return PARAMETERS; }
randoms		{ return RANDOMS; }
states		{ return STATES; }
times		{ return TIMES; }
variables	{ return VARIABLES; }

{HEX}		{ long int j = 0;
		  sscanf(((char *) yytext) + 2, "%lx", &j);
		  zzlval.val = double(j);
		  return NUM;
		}
{OCT}		{ long int j = 0;
		  sscanf((char *) yytext, "%lo", &j);
		  zzlval.val = double(j);
		  return NUM;
		}
{FLOAT}		{ sscanf((char *) yytext, "%lf", &zzlval.val); return NUM; }
Pi		{ zzlval.val = M_PI; return NUM; }
{ID} 		{
  		   symrec *s;
		   char *symbuf = new char[MAXSTRINGSIZE];
		   sscanf((char *) yytext, "%s", symbuf);
		   if (!(s = getsym (symbuf)))
		     s = putsym(symbuf, VAR, false);
		   delete [] symbuf;
		   zzlval.tptr = s;
		   return s->type;
		}
{STRING}	{
  		   zzlval.text = new char[MAXSTRINGSIZE];
		   /* Zap first `"` */
		   strcpy(zzlval.text, (const char *) yytext + 1);
		   /* Zap last `"' */
		   zzlval.text[strlen((const char *) yytext) - 2] = '\0';
		   return STRING;
		   // Don't forget to delete [] in the parser.
		}
"||"		{ return OR; }
"&&"		{ return AND; }
"=="		{ return EQ; }
"!="		{ return NE; }
">="		{ return GE; }
"<="		{ return LE; }

"+="		{ return PLUSE; }
"-="		{ return MINUSE; }
"*="		{ return TIMESE; }
"/="		{ return SLASHE; }
"%="		{ return PERCENTE; }
"<<"		{ return LL; }
">>"		{ return GG; }

"++"		{ return PLUSPLUS; }
"--"		{ return MINUSMINUS; }

"**"		{ return '^'; }

^{SHELLCMD}	{
		  FILE *f;
		  char line[MAXSTRINGSIZE];
		  f = popen((const char *) yytext + 1, "r");
		  while (fgets(line, MAXSTRINGSIZE, f)) {
		    printf(line);
		  }
		  pclose(f); }

{COMMENT}	{}
[ \t]+		{}
\\[ \t]*\n      {}

\n    		{ return (int) *yytext; }
.     		{ return (int) *yytext; }
%%
#ifdef __GNUC__
/*
 * This is just to get rid of stupid error message about yyunput not used.
 */
extern void never_called();

extern void 
never_called()
{
  yyunput(1, (char *)1);
#if 0
  // The latter 3 are for flex 2.4.6-2.4.7
  yy_push_state(1);
  yy_pop_state();
  yy_top_state();
#endif
}
#endif
