/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

// $Id: misc.cc,v 1.25 1994/08/24 09:58:29 bengt Exp $

// This file contains miscellaneous support routines

#include "externals.h"
#include "siteprefs.h"
#include "CIntegrator.h"

#define String XtString
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <stdlib.h>
#include <string.h>
#undef String
#include <unistd.h>

#ifdef RICCATI
#include "UserMatrix.h"
#endif

#define log if (do_logging) va_log_event

// Sets an integration routine.
extern void
select_algorithm(Integrator *alg)
{
  if (current_algorithm && current_algorithm->stochastic()
      && ! alg->stochastic())
    timestep = DEFAULT_TIMESTEP;
  if (alg->stochastic() && timestep == 0.0)
    timestep = default_stochastic_timestep;
  current_algorithm = alg;
  method_flag = current_algorithm->default_method_flag();
  log("switch algorithm (%s)", current_algorithm->name());
}

// Sets an integration routine, selected by name.
extern bool
select_algorithm(const char *name, bool critical)
{
  ContinuousIntegrator **algo;

  for (algo = algorithms; *algo; algo++)
    if (strcasecmp((*algo)->name(), name) == 0)
      break;

  if (! *algo)
    if (critical) 
      gnans_error(impossible, "Could not find DEFAULT algorithm ``%s''",
		  name);
    else
      gnans_error(non_fatal, "Could not find algorithm ``%s''", name);
  else
    select_algorithm(*algo);
  return bool(*algo);
}

extern bool
select_algorithm(const char *name)
{
  if (current_system.system_type != continuous) {
    gnans_error(non_fatal,
		"With the currently selected system you cannot select another algorithm.");
    return false;
  } else
    return select_algorithm(name, false);
}

// Selects an integrator, based on its index in the table.
extern void
select_algorithm(unsigned int index)
{
  select_algorithm(algorithms[index]);
}

// Sets a store format, selected by name.
extern bool
select_format(const char *name)
{
  unsigned int i;
  for (i = 0; store_format_table[i].write; i++)
    if (strcasecmp(store_format_table[i].name, name) == 0)
      break;
  if (! store_format_table[i].write) {
    gnans_error(non_fatal, "Could not find format %s", name);
    return false;
  } else {
    store_format = i;
    return true;
  }
}

/*
 * General purpose "finish" function. Arguments are a text message, and an
 * exit code.
 */
extern void 
die(const char *message, int errorcode)
{
  if (errorcode != lock_exists) {
    // Flush and close files
    // Should we write trajectories at exit or not? I don't know...
    // Let's write if append_trajectories is selected.
    if (append_trajectories)
      write_trajectories();
    close_log();
    unlink(GNANSLOCKFILENAME);
  }
  
  if (X)
    destroy_toplevel();
  if (*message)
    fprintf(stderr, "%s\n", message);
  if (gnansdebug > 1000) {
    fprintf(stderr, "debug is %d, dumping core as you requested.\n",
		   gnansdebug);
    abort();
  }
#ifndef DYNAMIC_LOADING
  fprintf(stderr, "To cleanup:\n"
		 "\trm -f xgnans f.gnans gstore.* gnans.log\n\r");
#endif
  exit(errorcode);
}

#ifdef MATLAB
#ifdef RICCATI
/**************** Preliminary stuff ****************/
extern "C" {
#include "mat.h"
}

extern bool
UserMatrix::matrix_init(char *nam, char *doc, bool store,
			 char *filenam, char *external_nam,
			 UserParameter **UserParameter_array)
{
  String name(nam);
  String documentation(doc);
  String filename(filenam);
  String external_name(external_nam);

  // load matrix and update dims
  MATFile *fp = matOpen(filename, "r");
  if (!fp) {
    gnans_error(non_fatal, "Cannot open file ``%s'' for reading",
		(char *) filename);
    return false;
  }

  double *imaginary_pointer;
  if (matGetFull(fp, external_name,
		 (signed int*) &number_of_rows,
		 (signed int*) &number_of_columns,
		 &data, &imaginary_pointer)) {
    gnans_error(non_fatal,
		"Cannot find (full) matrix named ``%s'' in file ``%s''",
		(char *) external_name, (char *) filename);
    return false;
  }

  if (imaginary_pointer) {
    gnans_error(non_fatal,
		"Matrix ``%s'' (in file ``%s'') is a complex matrix",
		(char *) external_name, (char *) filename);
    return false;
  }

  if (matClose(fp)) {
    return false;
    gnans_error(non_fatal, "Error occured when closing file ``%s''",
		(char *) filename);
  }

  // Now enter its elements as UserParameter's
  for (unsigned int i = 0; i < number_of_columns; i++) 
    for (unsigned int j = 0; j < number_of_rows; j++) {
      String element_name = name + "_" + dec(j+1) + "_" + dec(i+1);
      *(UserParameter_array++) =
	new UserParameter(element_name, documentation,
			   store, data[i*number_of_rows + j],
			   data + i*number_of_rows + j);
    }
  return true;
}
#endif
#endif
