/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

/* $Id: random.cc,v 1.6 1993/09/19 11:28:41 bengt Exp $ */


#ifdef __GNUG__
#pragma implementation
#endif

#include "externals.h"
#ifndef GNANS
#include <math.h>
#endif
#include "random.h"

#ifdef GNANS

static void 
error(char *str)
{
  gnans_error(non_fatal, str);
}

#else				// ! GNANS

#if defined(TEST_U) || defined(TEST_N) || defined(TESTSAVE)
#include <stdlib.h>
#include <stream.h>
#include <stdio.h>

static void 
error(char *str)
{
  fprintf(stderr, "%s\n", str);
  abort();
}
#endif

#endif				// ! GNANS

#if !defined(BM) && !defined(PM)
#define INVPHI
#endif

#if 0
#define MAX_SAVES 32

/* Feedback polynomial: x^63 + x + 1  */

static const int rand_deg = 63;
static const int rand_sep = 1;
#endif

static const long int virgin_state[rand_deg] =
{ 
  0x9a319039, 0x32d9c024, 0x9b663182, 0x5da1f342, 
  0xde3b81e0, 0xdf0a6fb5, 0xf103bc02, 0x48f340fb, 
  0x7449e56b, 0xbeb1dbb0, 0xab5c5918, 0x946554fd, 
  0x8c2e680f, 0xeb3d799f, 0xb11ee0b7, 0x2d436b86, 
  0xda672e2a, 0x1588ca88, 0xe369735d, 0x904f35f7, 
  0xd7158fd6, 0x6fa6f051, 0x616e6b96, 0xac94efdc, 
  0x36413f93, 0xc622c298, 0xf5a42ab8, 0x8a88d77b, 
  0xf5ad9d0e, 0x8999220b, 0x27fb47b9, 0x27fb47b9,
  0x9a319039, 0x32d9c024, 0x9b663182, 0x5da1f342, 
  0xde3b81e0, 0xdf0a6fb5, 0xf103bc02, 0x48f340fb, 
  0x7449e56b, 0xbeb1dbb0, 0xab5c5918, 0x946554fd, 
  0x8c2e680f, 0xeb3d799f, 0xb11ee0b7, 0x2d436b86, 
  0xda672e2a, 0x1588ca88, 0xe369735d, 0x904f35f7, 
  0xd7158fd6, 0x6fa6f051, 0x616e6b96, 0xac94efdc, 
  0x36413f93, 0xc622c298, 0xf5a42ab8, 0x8a88d77b, 
  0xf5ad9d0e, 0x8999220b, 0x27fb47b9
  };

void
rng::srandom(long int x)
{
  register long int i;
  state[0] = x;
  for (i = 1; i < rand_deg; ++i)
    state[i] = (1103515145 * state[i - 1]) + 12345;
  for (i = 0; i < 10 * rand_deg; ++i)
    random();
#ifndef INVPHI
  one_left = 0;
#endif
}

void
rng::setstate(long int *arg_state)
{
  state = arg_state;
  fptr = &randtbl[rand_sep];
  rptr = randtbl;
  end_ptr = &randtbl[sizeof(randtbl) / sizeof(randtbl[0])];
#ifndef INVPHI
  one_left = 0;
#endif
}

long int *
rng::getstate()
{
  return state;
}

int
rng::save_state(int index)
{
  long int *ptr = rptr;
  if (index >= MAX_SAVES || index < 0) {
    error("Erroneous index for saving state.");
    return -1;
  }
  if (!saved_states[index])
    saved_states[index] = new long int[rand_deg];
  for (int i = 0; i <= rand_deg - 1; i++) {
    saved_states[index][i] = *ptr;
    if (++ptr >= end_ptr)
      ptr = state;
  }
  return index;
}
  
int
rng::save_state()
{
  while (number_of_saved_states < MAX_SAVES
	 && saved_states[number_of_saved_states])
    number_of_saved_states++;
  if (number_of_saved_states >= MAX_SAVES) {
    error("Maximal number of saved states reached.");
    return -1;
  }
  return save_state(number_of_saved_states++);
}

int
rng::restore_state(int index)
{
  if (index > MAX_SAVES || index < 0) {
    error("Erroneous state index.");
    return 0;
  }
  if (!saved_states[index]) {
    error("No state saved with that index.");
    return 0;
  }
  for (int i = 0; i <= rand_deg - 1; i++)
    randtbl[i] = saved_states[index][i];
  setstate(randtbl);
  return 1;
}
      
rng::rng()
{
  int i;
  invmax      = 0.46566128730773925781e-9; /* 1/2^31 */
  invmax_2    = 0.93132257461547851563e-9; /* 2/2^31 */
  invmax_2_pi = 0.29258361585343193622e-8; /* 2 Pi/2^31 */
  number_of_saved_states = 0;
#ifndef INVPHI
  one_left = 0;
#endif
  for (i = 0; i <= rand_deg - 1; i++)
    randtbl[i] = virgin_state[i];
  for (i = 0; i <= MAX_SAVES - 1; i++)
    saved_states[i] = NULL;
  setstate(randtbl);
  overhead = 0;
}

rng::rng(unsigned long int n)
{
  for (int i = 0; i <= MAX_SAVES - 1; i++)
    saved_states[i] = NULL;
  setstate(randtbl);
  srandom(n);
}

rng::~rng()
{
  for (int i = 0; i <= MAX_SAVES - 1; i++)
    if (saved_states[i])
      delete saved_states[i];
}

void
rng::set_overhead(int n)
{
  overhead = n;
}

/****************************************************************/
#if defined(TEST_U) || defined(TEST_N)
extern int
main(int argc, char *argv[])
{
  register int i;
  int limit;
  double x;
  rng rand;
  if (argc != 2) {
    fprintf(stderr, "%s: One argument (number of samples) required\n", argv[0]);
    exit(1);
  }
  limit = atoi(argv[1]);

  for (i = 1; i <= limit; i++) {
#ifdef TEST_U
    x = rand.uniform();
#else
    x = rand.normal();
#endif
#ifdef PRINT
    cout << x << "\n";
#endif
  }
  exit(0);
}
#endif

#ifdef TESTSAVE
#define SAMPLES 10
extern int
main()
{
  rng rand;
  int i, first, second;

  first = rand.save_state();
  for (i = 1; i <= SAMPLES; i++)
    cout << rand.uniform() << "\n";
  printf("-----\n");
  second = rand.save_state();
   for (i = 1; i <= SAMPLES; i++)
     cout << rand.uniform() << "\n";
   printf("-----\n");
 
  rand.restore_state(first);
  for (i = 1; i <= SAMPLES; i++)
    cout << rand.uniform() << "\n";
} 
#endif

#ifdef static
#undef static
#endif
