/*
 Copyright  1992, 1993 Bengt Martensson and Institute for Dynamical Systems
 Copyright  1985, 1986, 1987, 1990 Free Software Foundation, Inc.


This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

// $Id: signals.cc,v 1.19 1994/08/19 18:38:31 bengt Exp $

/*
 * Signal handlers and setup routine. The fatal signal handler is derived
 * from Gnu Emacs.
 *
 * This file may contain some conditional compilation...
 */

/*
 * System calls like write (in write_trajectories()) are probably not
 * protected the way they should be.
 */

#include "externals.h"
#include <signal.h>
#include <errno.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_SYS_TIME_H 
#include <sys/time.h>
#endif

#ifdef IEEE_EXCEPTIONS
#include <floatingpoint.h>
#endif

/*#define SIGNAL_CONSIDERED_HARMFUL*/
/*
 * Note: No calls causing X calls are permitted from signal handlers.
 * In particular, we temporarily disable X interface during error messages.
 */

typedef RETSIGTYPE (*sighandler_t)(int);
static void
install_signal_handler(const int thesignal, sighandler_t handler)
{
#ifdef SIGNAL_CONSIDERED_HARMFUL
  struct sigaction sa;
  sa.sa_handler = handler;
  sigaction(thesignal, &sa, NULL);
#else
  signal(thesignal, handler);
#endif
}

static void
install_signal_handler_ignore(const int thesignal)
{
#ifdef SIGNAL_CONSIDERED_HARMFUL
  struct sigaction sa;
  sa.sa_handler = (sighandler_t) SIG_IGN; // Awful that this cast is needed...
  sigaction(thesignal, &sa, NULL);
#else
  signal(thesignal, SIG_IGN);
#endif
}

static void
install_signal_handler_default(const int thesignal)
{
#ifdef SIGNAL_CONSIDERED_HARMFUL
  struct sigaction sa;
  sa.sa_handler = (sighandler_t) SIG_DFL; // Awful that this cast is needed...
  sigaction(thesignal, &sa, NULL);
  sigset_t mask;
  sigemptyset(&mask);
  sigaddset(&mask, thesignal);
  sigprocmask(SIG_UNBLOCK, &mask, NULL);
#else
  signal(thesignal, SIG_DFL);
#endif
}

/*
 * If SIGINT, stop run if running from terminal, otherwise quit program.
 */
static RETSIGTYPE
sigint_handler(int /*not_used*/)
{
  install_signal_handler_ignore(SIGINT);
  if (X)
    gnans_error(fatal, "SIGINT received, quitting");

  /* Yes, this code is somewhat redundant. */
  if (running_integration || running_script) {
    bool Xsave = X;
    X = false;
    gnans_error(message, "SIGINT received, stopping");
    X = Xsave;
  }
  running_integration = false;
  if (X && workprocs && running_script)
    abort_script = true;
  install_signal_handler(SIGINT, &sigint_handler);
  return;
}

/*
 * If SIGUSR1 (was SIGHUP), display variables.
 */
static RETSIGTYPE
display_handler(int /*not_used*/)
{
  bool Xsave = X;
  X = false;
  install_signal_handler_ignore(SIGUSR1);
  gnans_error(nonerror_message, "SIGUSR1 received, displaying variables");
  Display_vars(0);
  install_signal_handler(SIGUSR1, &display_handler);
  X = Xsave;
  return;
}

// I don't dare to do a popup error message from the signal handler.

/*
 * Unconditionally quit on SIGTERM or SIGHUP.
 */
static RETSIGTYPE
sigterm_handler(int /*not_used*/)
{
  bool Xsave = X;
  X = false;
  gnans_error(fatal, "SIGTERM or SIGHUP received, quitting");
  X = Xsave;
  return;
}

/*
 * Alarm is used by the workprocs. The signal handler just sets a flag.
 */
static RETSIGTYPE
sigalrm_handler(int /*not_used*/)
{
  install_signal_handler_ignore(SIGALRM);
  if (trace_workproc)
    gnans_error(debug, "SIGALRM detected");
  if (workprocs) {
    has_been_alarmed = true;
    if (running_integration || running_script)
      set_alarm();
  }
  return;
}

/*
 * Sets the alarm to TIMESLICE microseconds, which will be the interval
 * at which user inputs are detected.
 */
extern void
set_alarm()
{
  struct itimerval timer;
  if (!X)
    return;

  timer.it_value.tv_sec = (int) (TIMESLICE/1000000.0);
  timer.it_value.tv_usec = TIMESLICE - 1000000 * timer.it_value.tv_sec;
  timer.it_interval.tv_sec = 0;
  timer.it_interval.tv_usec = 0;
  if (setitimer(ITIMER_REAL, &timer, NULL))
    gnans_error(impossible, "Error %d in setitimer, secs = %d, usecs = %d",
		errno, timer.it_value.tv_sec, timer.it_value.tv_usec);
  install_signal_handler(SIGALRM, &sigalrm_handler);
}

/*
 * Cancels previously set alarm.
 */
extern void
cancel_alarm()
{
  install_signal_handler_ignore(SIGALRM);
}

/*
 * Called when the workprocs switch is switched. Just installs the correct
 * signal handler.
 */
extern void
update_workproc_status()
{
  if (workprocs)
    install_signal_handler(SIGALRM, &sigalrm_handler);
  else {
    install_signal_handler_ignore(SIGALRM);
    has_been_alarmed = false;
  }
}

/****************************************************************/
/*
 * The following two functions are derived from Gnu Emacs (emacs.c)
 */

/* Signal code for the fatal signal that was received */
static int fatal_error_code;

/* true if handling a fatal error already */
static bool fatal_error_in_progress;

/* Handle bus errors, illegal instruction, etc. */
static RETSIGTYPE
fatal_error_signal_handler(int sig)
{
  unlink(GNANSLOCKFILENAME);
  fatal_error_code = sig;
  install_signal_handler_default(sig);
  
  /* If fatal error occurs in code below, avoid infinite recursion.  */
  if (fatal_error_in_progress)
    kill(getpid(), fatal_error_code);
  fatal_error_in_progress = true;

  if (sig != SIGTERM)
    gnans_error(debug, "Got fatal signal (%d)", sig);
  va_log_event("Got fatal signal (%d)", sig);
  emergency_dump();		/* Closes the log & removes lock */

  /* Signal the same code; this time it will really be fatal.  */
  kill(getpid(), fatal_error_code);
  return;
}

#ifdef IEEE_EXCEPTIONS
/*
 * Floating point exception handlers
 */
static void
fp_overflow_handler(int sig, int code, struct sigcontext *scp, char *addr)
{
  if (development && gnansdebug > 1000)
    gnans_error(impossible,	// Dump core
		"IEEE exception code %x (overflow) occurred at pc %X \n",
		code,scp->sc_pc);
  else {
    gnans_error(signal_handler_message, "IEEE overflow exception");
    ieee_flags("clear", "exception", "overflow", NULL);
    
    // Stop ongoing work
    running_integration = false;
    floating_point_error_occured = true;
    if (running_script)
      abort_script = true;
    // throw(...)
  }
    
}

void
fp_invalid_handler(int sig, int code, struct sigcontext *scp, char *addr)
{
  if (development && gnansdebug)
    gnans_error(impossible,
		"IEEE exception code %x (Invalid) occurred at pc %X \n",
		code,scp->sc_pc);
  else {
    gnans_error(signal_handler_message, "IEEE invalid exception");
    ieee_flags("clear", "exception", "invalid", NULL);

    running_integration = false;
    floating_point_error_occured = true;
    if (running_script)
      abort_script = true;
    // throw(...);
  }
}
#endif

/*
 * Initial setup of the signal handlers. (Derived from Gnu Emacs.)
 */
extern void
signal_setup()
{
  install_signal_handler(SIGUSR1, &display_handler);
  install_signal_handler(SIGINT, &sigint_handler);
  install_signal_handler(SIGALRM, &sigalrm_handler);
  install_signal_handler(SIGTERM, &sigterm_handler);
  install_signal_handler(SIGHUP, &sigterm_handler);

  install_signal_handler(SIGQUIT, &fatal_error_signal_handler);
  install_signal_handler(SIGILL, &fatal_error_signal_handler);
  install_signal_handler(SIGTRAP, &fatal_error_signal_handler);
  install_signal_handler(SIGIOT, &fatal_error_signal_handler);
#ifdef SIGEMT
  install_signal_handler(SIGEMT, &fatal_error_signal_handler);
#endif
  //install_signal_handler(SIGFPE, &fp_handler);
  ///*fatal_error_signal_handler
  install_signal_handler(SIGBUS, &fatal_error_signal_handler);
  install_signal_handler(SIGSEGV, &fatal_error_signal_handler);
#ifdef SIGSYS
  install_signal_handler(SIGSYS, &fatal_error_signal_handler);
#endif
#ifdef SIGXCPU
  install_signal_handler(SIGXCPU, &fatal_error_signal_handler);
#endif
#ifdef SIGXFSZ
  install_signal_handler(SIGXFSZ, &fatal_error_signal_handler);
#endif /* SIGXFSZ */

#ifdef AIXV3
  install_signal_handler(SIGDANGER, &fatal_error_signal_handler);
  //install_signal_handler(20, &fatal_error_signal_handler);
  install_signal_handler(21, &fatal_error_signal_handler);
  install_signal_handler(22, &fatal_error_signal_handler);
  install_signal_handler(23, &fatal_error_signal_handler);
  install_signal_handler(24, &fatal_error_signal_handler);
  install_signal_handler(SIGAIO, &fatal_error_signal_handler);
  install_signal_handler(SIGPTY, &fatal_error_signal_handler);
  install_signal_handler(SIGIOINT, &fatal_error_signal_handler);
  install_signal_handler(SIGGRANT, &fatal_error_signal_handler);
  install_signal_handler(SIGRETRACT, &fatal_error_signal_handler);
  install_signal_handler(SIGSOUND, &fatal_error_signal_handler);
  install_signal_handler(SIGMSG, &fatal_error_signal_handler);
#endif /* AIX */

#ifdef IEEE_EXCEPTIONS
  sigfpe_handler_type inexact, division, underflow;
  inexact = (sigfpe_handler_type) ieee_handler("get", "inexact", NULL);
  division = (sigfpe_handler_type) ieee_handler("get", "division", NULL);
  underflow = (sigfpe_handler_type) ieee_handler("get", "underflow", NULL);
  
  ieee_handler("set", "inexact", inexact);
  ieee_handler("set", "division", division);
  ieee_handler("set", "underflow", underflow);
  ieee_handler("set", "overflow", (sigfpe_handler_type) fp_overflow_handler);
  ieee_handler("set", "invalid", (sigfpe_handler_type) fp_invalid_handler);
#endif
}
