/* -*-c-*- */
/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

/* $Id: scanner.l,v 1.21 1994/11/01 12:53:43 bengt Exp bengt $ */

%{
/*
 * This scanner works in one of three modes:
 * INITIAL	This is start mode. For declarations of states etc.
 * eqns		Equations (in the sense of gnans) are parsed.
 * raw		Dumb mode. Uninterpreted C-code is just passed thru.
 */

#include <config.h>
#include "parser.h"
#include "translator.h"
#include <string.h>
#include <math.h>
#include <utillib.h>

#ifndef M_PI			/* if __STRICT_ANSI__ is in effect... */
#define M_PI            3.14159265358979323846
#endif

#define yytext_ptr yytext
/* This is to keep track of what state the comment was called */
static int prev_state;

/* Keep track of the nesting level within raw C code */
static int nesting_depth = 0;

static void
inc_lineno() {
  if (trace_lineno)
    fprintf(stderr, "End of line# %d\n", lineno);
  lineno++;
}

/* Returns the number of newlines found in a string */
static int 
number_of_newlines(const char *str)
{
  return !str ? 0 :
    strchr(str, '\n') ? number_of_newlines(strchr(str, '\n') + 1) + 1 : 0;
}

#undef YY_INPUT
#define YY_INPUT(buf,result,max_size) my_yy_input(buf, result, max_size)

static void
my_yy_input(char *buf, int &result, int max_size)
{
  if (! fgets((char *) buf, max_size, yyin))
    result = YY_NULL;
  else {
    result = strlen((char *) buf);
    delete [] current_line;
    current_line = new char[result + 1];
    strcpy(current_line, (char *) buf);
  }
}
%}

%x raw
%s eqns
%x comment
%x C_block

%option noyywrap

DIGITS	[0-9]+
INT	{DIGITS}
FLOAT 	({DIGITS}?("."{DIGITS})?)|({DIGITS}("."{DIGITS})?[eE][+-]?{DIGITS})|(Inf(inity)?)
ID	[a-zA-Z][a-zA-Z0-9_]*
ID8	[a-zA-Z][a-zA-Z0-9_\320-\377]*
STRING 	\"[^\"]*\"

%%
<INITIAL>CONTINUOUS		{ return CONTINUOUS; }
<INITIAL>DISCRETE		{ return DISCRETE; }
<INITIAL>HYBRID			{ return HYBRID; }
<INITIAL>DAS			{ return DAS; }
<INITIAL,eqns>RICCATI		{ return RICCATI; }
<INITIAL>TIME			{ return TIME; }
<INITIAL>EVENT			{ return EVENT; }
<INITIAL>SYSTEM			{ return SYSTEM; }

<INITIAL,eqns>STATE		{ return STATE; }
<INITIAL,eqns>INTEGER[ \t]*STATE { return INTEGERSTATE; }
<INITIAL,eqns>PARAMETER		{ return PARAMETER; }
<INITIAL,eqns>FUNCTION		{ return FUNCTION; }
<INITIAL,eqns>VARIABLE		{ return VARIABLE; }
<INITIAL,eqns>UNIFORM		{ return UNIFORM; }
<INITIAL,eqns>NORMAL		{ return NORMAL; }
<INITIAL,eqns>WHITE		{ return WHITE; }
<INITIAL,eqns>NOISE		{ return NOISE; }
<INITIAL,eqns>WIENER		{ return WIENER; }
<INITIAL,eqns>PROCESS		{ return PROCESS; }
<INITIAL,eqns>MATRIX		{ return MATRIX; }
<INITIAL,eqns>VECTOR		{ return VECTOR; }
<INITIAL,eqns>C-IDENTIFIER	{ return C_IDENTIFIER; }
<INITIAL,eqns>FILEREAD		{ return FILEREAD; }
<INITIAL,eqns>DeBuG             { gnans_error(impossible, "Debugging..."); }

<INITIAL,raw,eqns>AT[ \t]*TIME	{ BEGIN(eqns);
				  return ATTIME; }
<eqns>d/"("			{ return (int) *yytext; }
<eqns>q/"("			{ return (int) *yytext; }
<INITIAL,eqns>"*""*"		{ return (int) '^'; }

<INITIAL,eqns>==                { return EQ; }
<INITIAL,eqns>"!"=              { return NE; }
<INITIAL,eqns>"<"=              { return LE; }
<INITIAL,eqns>">"=              { return GE; }
<INITIAL,eqns>"|""|"            { return OR; }
<INITIAL,eqns>"&""&"            { return AND; }
<INITIAL,eqns>"<""<"            { return LL; }
<INITIAL,eqns>">"">"            { return GG; }
<INITIAL,eqns>"+""+"            { return PLUSPLUS; }
<INITIAL,eqns>"-""-"            { return MINUSMINUS; }


<eqns>"{"			{ nesting_depth = 1;
				  if (trace_lineno)
				    fprintf(stderr, "Nesting = %d\n",
					    nesting_depth);
				  BEGIN(C_block);
				  *current_block = "#line " + number_String(lineno) + " \"" + input_filename + "\"\n";
				  *current_block += yytext; }
<C_block>"{"			{ nesting_depth++;
				  if (trace_lineno)
				    fprintf(stderr, "Nesting = %d\n",
						   nesting_depth);
				  *current_block += yytext; }
<C_block>"}"			{ nesting_depth--;
				  if (trace_lineno)
				    fprintf(stderr, "Nesting = %d\n",
					    nesting_depth);
				  *current_block += yytext;
				  if (nesting_depth == 0) {
				    BEGIN(eqns);
				    return C_BLOCK;
				  }
				}
<C_block>\n			{ lineno++;
				  *current_block += yytext; }
<C_block>.			{ *current_block += yytext; }

 /* These are only for casts */
<INITIAL,eqns>int		{ return INTEGER; }
<INITIAL,eqns>double		{ return DOUBLE; }

 /* INT must take preference before FLOAT */
<INITIAL,eqns>{INT}		{ sscanf((char *) yytext, "%d",
					 &zzlval.integer);
				  return INT; }
<INITIAL,eqns>{FLOAT}		{ sscanf((char *) yytext, "%lf",
					 &zzlval.value);
				  return FLOAT; }
<INITIAL,eqns>Pi		{ zzlval.value = M_PI; return FLOAT; }
<INITIAL,eqns>{ID}		{ zzlval.text = new char[strlen(yytext) + 1];
				  strcpy(zzlval.text, (const char *) yytext);
				  return ID; }
<INITIAL,eqns>{ID8}		{ zzlval.text = new char[strlen(yytext) + 1];
				  strcpy(zzlval.text, (const char *) yytext);
				  return ID8; }
<INITIAL,eqns>{STRING}		{ zzlval.text = new char[strlen(yytext) + 1];
				  strcpy(zzlval.text, (const char *) yytext);
				  lineno +=
                                    number_of_newlines((char *) yytext);
				  return STRING; }

<INITIAL,eqns>C-DECLARATION 	{ current_block = &c_declarations;
				  BEGIN(raw); }

^[ \t]*#.*\n     		{ inc_lineno();
                                  fprintf(yyout, "%s", yytext);
				  if (strncmp(yytext, "#line ", 6))
				      cpp_calls++; }
<INITIAL,eqns>"//".*\n	 	{ inc_lineno(); }
<INITIAL>"/*"			{ prev_state = INITIAL;
				  BEGIN(comment); }
<eqns>"/*"			{ prev_state = eqns;
				  BEGIN(comment); }
<raw>"/*"			{ prev_state = raw;
				  BEGIN(comment); }
<comment>[^*\n]*		{ }
<comment>"*"+[^*/\n]*		{ }
<comment>\n			{ inc_lineno(); }
<comment>"*"+"/"		{ BEGIN(prev_state); }

<INITIAL,eqns>[ \t]+		{}
<INITIAL,eqns>\n		{ inc_lineno();}

.	 			{ return (int) *yytext; }
<raw>\n				{ inc_lineno();
				  *current_block += yytext;}
<raw>.				{ *current_block += yytext; }
%%
/*
 * This is just to get rid of stupid error message about yyunput not used ...
 */
extern void never_called();
extern void 
never_called()
{
  yyunput(1, (char *)1);
  // The latter 3 are for flex 2.4.6-2.4.7
#if 0
  yy_push_state(1);
  yy_pop_state();
  yy_top_state();
#endif
}
