/*
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

/* $Id: support.cc,v 1.17 1994/08/17 18:57:46 bengt Exp $ */

/* This file defines a bunch of utility functions for the translator. */

/****************************************************************/
#include <config.h>
#include "translator.h"

/**************** Rotines for lists ****************/

/* A new node as first element */
extern symrec *
putsym(const char *sym_name, sym_type type)
{
  symrec *ptr = new symrec(sym_name, type, sym_table);
  sym_table = ptr;
  return ptr;
}

/* Fetch the symbol with a given name */
extern symrec *
getsym(const char *sym_name)
{
  symrec *ptr;
  for (ptr = sym_table; ptr; ptr = ptr->next)
    if (!strcmp(ptr->name, sym_name))
      return ptr;
  return NULL;
}

/* Enter a new symbol in the list, provided it is not there already */
extern symrec *
put_unique_sym(const char *name, sym_type type)
{
  symrec *s = getsym(name);
  if (s)
    if (s->type == Keyword) {
      gnans_error(semantic, "The name `%s' is reserved", name);
    } else
      gnans_error(semantic, "The name `%s' is already in use", name);
  else
    s = putsym(name, type);
  return s;
}
    
/* Adds a dependency to the currently defined equation */
extern void 
add_dependency(const char *name)
{
  deprec *d, *dptr = new deprec;
  symrec *ptr = new symrec;
  if ((ptr = getsym(name))) {
    if (ptr->type == Parameter ||
	ptr->type == Parameter_vector ||
	ptr->type == Parameter_matrix)
      return;
    for (d = current_dependencies; d; d = d->next) {
      if (d->symbol == ptr) {
	return;			/* Already there */
      }
    }
    dptr->symbol = ptr;
    dptr->next = current_dependencies;
    current_dependencies = dptr;
  } else
    gnans_error(semantic,
		"`%s' (in the right hand side of last equation) undeclared",
		name);
}

extern void 
dependency_delete(const char *name)
{
  symrec *p;
  deprec *d, **prev;
  for (p = sym_table; p; p = p->next) {
    prev = &p->dependencies;
    for (d = p->dependencies; d; d = d->next) {
      if (!d->symbol || !strcmp(d->symbol->name, name)) {
	*prev = d->next;
      } else {
	prev = &(d->next);
      }
    }
  }
}

/* Delete an entry from the symtab */
extern void 
symtab_delete(const char *name)
{
  symrec *p, **prevptr;
  prevptr = &sym_table;
  for (p = sym_table; p; p = p->next) {
    if (!strcmp(p->name, name)) {
      *prevptr = p->next;
    } else
      prevptr = &(p->next);
  }
  dependency_delete(name);
}

/* Delete all entries of a certain type from the symtab */
extern void 
symtab_delete_type(sym_type type)
{
  symrec *p, **prevptr;
  if (deps_debug)
    fprintf(stderr, "Zapping all entries of type %d\n", type);
  prevptr = &sym_table;
  for (p = sym_table; p; p = p->next) {
    if (p->type == type) {
      *prevptr = p->next;
      dependency_delete(p->name);
    } else
      prevptr = &(p->next);
  }
}

/* Check that all symbols which are states, functions, or variables */
/* are assigned */
extern void 
assignment_check()
{
  symrec *ptr;
  for (ptr = sym_table; ptr != (symrec *) 0; ptr = (symrec *)ptr->next) {
    switch(ptr->type) {
    case State:
    case State_vector:
    case State_matrix:
    case Function:
    case Variable:
      if (!ptr->equation) {
	lineno--;
	gnans_error(recoverable, "`%s' has no equation", ptr->name);
      }
      break;
    default:
      break;
    }
  }
}

/**************** Error handling ****************/
extern void 
zzerror(const char */*s*/)
{
  gnans_error(parse, "");
}
