/* -*- c++ -*-
 Copyright (C) 1992 Bengt Martensson and Institute for Dynamical Systems.

This file is part of Gnans.

Gnans is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

Gnans is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Gnans; see the file COPYING.  If not, write to the 
Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. */

/* $Id: translator.h,v 1.21 1994/08/17 18:57:46 bengt Exp $ */

#include <stdio.h>		// for FILE*
#include <string>
#include <Enums.h>
#include <string.h>

/**************** Compilation parameters ****************/
/* These are only used for sizing data structures */
/* Should be "large" but not "gigantic" */

#define BIGSTRINGSIZE 100000	/* Only used in va_funcs.cc */
#define MAXSTATES 1000
#define SMALLTABLESIZE 5
#define RANDOM_INT_LENGTH	64

#define OUTPUT_ENDING ".cc"

/**************** Macros ****************/
#define NODESCRIPTION "*** No description available ***"
#define STARTINDEX 0
#define STOPINDEX  1
#define STOREINDEX 2
#define ACCEPTEDINDEX 3

/**************** Data structures ****************/
typedef enum {			/* Type of symbol */
  State,
  State_vector,
  State_matrix,
  Integer_State,
  Parameter,
  Parameter_vector,
  Parameter_matrix,
  Time_type,
  Function,
  Variable,
  Wiener_coeff,
  Uniform_Noise,
  Normal_Noise,
  Wiener_Process,
  C_Identifier,
  Keyword,			/* Keyword in C or C++, reserved */
  Formal_Parameter,		/* Formal parameter in function definition */
  Defun				// User defined function
} sym_type;

class symrec {			/* Entry in the symbol table */
public:
  char *name;
  sym_type type;
  int a_priori_timeindex; /* Was this defined in e.g. AT TIME START? */

  /* For states: The RHS of the defining equation
     for wieners and times: scratch, the coeff associated with this wiener
     in the current equation*/
  char *equation;

  unsigned int line_no_of_equation; /* Line # of the (end of) the equation */
  struct symrec *next;
  struct symrec *parent;	/* "Parent" (used for dependencies) */
  struct deprec *dependencies;	/* Pointer to a list of symbols of which */
				/* this symbol depends on */
  symrec(const char *sym_name, sym_type t, symrec *nxt)
    {
      name = new char[strlen((char*) sym_name) + 1];
      strcpy(name, sym_name);
      type = t;
      next = nxt;
      a_priori_timeindex = -1;
      equation = NULL;
      dependencies = NULL;
      parent = NULL;
    }
  symrec()
    {
      name = NULL;
      next = NULL;
      a_priori_timeindex = -1;
      equation = NULL;
      dependencies = NULL;
      parent = NULL;
    }
  ~symrec() 		{ delete [] name; delete [] equation; }
};

class deprec {			/* Entry in the dependency list */
public:
  struct symrec *symbol;
  struct deprec *next;
};

class translator_par {
  char *name;
public:
  char *documentation;
  char *filename;
  char *external_name;
  double initial;
  operator char*()	                { return name; }
  char * operator=(char *nam)	        { name = nam; return name; }
  char * operator=(std::string s)	        { name = new char[s.length() + 1];
					  strcpy(name, s.c_str());
					  return name; }
  char * operator+(unsigned int n)      { return name + n; }
};
  
class riccati_data_class {
 public:
  symrec *A11, *A12, *A21, *A22;
  unsigned int nnq, nnp;

//  dX/dt = A21(t) + A22(t)*X -X*A11(t) - X*A12(t)*X

//               nnq     nnp     
//             |             |
//        nnq{ |A(t)  | A(t) |
//             | 11   |  12  |  
//  A(t):=     | -----|----  |  
//             |      |      |  
//        nnp{ |A(t)  | A(t) |  
//             | 21   |  22  |  
//
};

/***************** Global variables *****************/
extern FILE *zzin, *zzout, *outputfile;
extern unsigned int number_of_states;	/* Number of states in the system */
extern unsigned int number_of_state_vectors; /* Number of state vectors in the system */
extern unsigned int number_of_state_matrices; /* Number of state matrices in the system */
extern unsigned int number_of_parameters; /* ... */
extern unsigned int number_of_parameter_vectors; /* ... */
extern unsigned int number_of_parameter_matrices; /* ... */
extern unsigned int number_of_functions;
extern unsigned int number_of_variables;
extern unsigned int number_of_times;
extern unsigned int number_of_randoms;
extern unsigned int number_of_uniforms;
extern unsigned int number_of_normals;
extern unsigned int number_of_wieners;

extern unsigned int lineno;	/* Line number in input file during parsing */
extern unsigned int function_calls; /* # of function calls in equations */
extern unsigned int cpp_calls;	    /* # of cpp usages */
extern SystemType system_type;	    /* continuous, discrete... */
extern std::string system_name;	    /* The name of the current system */
extern std::string input_filename;	    /* Name of current input file */
extern std::string output_filename;	    /* Name of current output file */

extern char *current_line;

/* Pointer to the buffer presently being written */
extern std::string *current_block;
extern std::string time_variable_name;
extern std::string c_declarations;
extern translator_par states[];		/* Array of names of states */
extern translator_par state_vectors[];
extern translator_par state_matrices[];
extern translator_par functions[];	/* ... */
extern translator_par parameters[];
extern translator_par parameter_vectors[];
extern translator_par parameter_matrices[];
extern translator_par variables[];
extern translator_par randoms[];
extern translator_par uniforms[];
extern translator_par normals[];
extern translator_par wieners[];
extern std::string times_names[];
extern std::string times_buffer[];	/* Each buffer contains a block of */
				/* uninterpreted C code  */

extern unsigned int translator_debug; /* Debugging code */
extern bool deps_debug;		/* Debugging the sorting */
extern bool trace_mallocs;	/* Tracing malloc calls */
extern bool trace_lineno;	/* Tracing line numbers during parsing */
extern symrec *sym_table;	/* Symbol table */
extern riccati_data_class riccati_data;
extern deprec *current_dependencies; /* Dependencies of currently parsed */
				     /* symbol */
extern const char *version_string;   /* Version number as string */

extern unsigned int errors_occured; // errors detected, don't produce output.

/**************** Friends of yacc/lex ****************/
#ifndef FLEX_SCANNER
/* If using an old flex version, should possibly be unsigned char */
extern char *zztext;
extern int zzleng;
extern int zzlex();
extern int zzparse();

extern FILE *zzin, *zzout;
extern int zz_flex_debug;
extern int zzdebug;
#endif

/**************** External support functions ****************/
extern char *stringify(const char *format, ...);
extern void gnans_error(severity sev, const char *format, ...);
extern symrec *getsym(const char *sym_name);
extern symrec *putsym(const char *sym_name, sym_type type);
extern symrec *put_unique_sym(const char *sym_name, sym_type type);
extern void add_dependency(const char *name);
extern void assignment_check();
extern void dependency_delete(const char *name);
extern void remove_crossdep(const char *name);
extern void symtab_delete(const char *name);
extern void symtab_delete_type(sym_type type);
extern void zzerror(const char *s);
